//#include <gsl/gsl_math.h>
#include "decs.h"

/* -------------------------------------------------------------------------- */

/* == parameters (limit to this subroutine) == */

/* parameter for non-thermal model */
#define NONTHERMAL_p 3.5
#define NONTHERMAL_eta 0.015

/* -------------------------------------------------------------------------- */

/* == options == */

/* Set to 1 if emiss.c is used outside bothros
 * Set to 0 if emiss.c is used with bothros */
#define STAND_ALONE_EMISS_SUBROUTINE (1)

/* condition for switching to T2 for the hybrid or combined equations */
//#define CONDITION_T2 (log10(Tdl)>0.5*(log10(nuratio)-1.)+0.2)
#define CONDITION_T2 (lg_Tdl>0.5*(lg_nuratio-1.)+0.2)

/* condition for switching to EX for the combined equations */
#define CONDITION_EX (lg_nuratio>11.||lg_Tdl<-0.5)

/* -------------------------------------------------------------------------- */

/* == Description of this file ==
 *  Calculate the specific emissivity and absorption coefficient of electron.
 *  The input variables and outputs of the jnu() and alphanu() are in cgs
 *  units. The gas is assumed to be neutral plasma with H^+ and e^- only (a
 *  reasonable assumption at high temperature because all the Helium and 
 *  metal nuclei are broken down to H^+).
 *
 * == About parameter EMISSION_EQ ==
 * == Bremsstrahlung ==
 *  B0EQ : PadI Eq.(6.227)
 * == Compton ==
 *  C0EQ : PadI Eq.(6.76)
 * == Thermal Synchrotron ==
 *  T0EQ : Pacholczyk70 Eq.(3.54), with I(x_M) set by using T0EQ_OPTION
 *  T1EQ : Wardzinski00 Eq.(7)-(10)
 *  T2EQ : Baring88 Eq.(10), see also Brainerd87 Eqs.(39), (40) and (45)
 *  T3EQ : Wardzinski00 Eq.(6)
 *  T4EQ : Baring88 Eq.(10), see also Brainerd87 Eqs.(39), (40) and (45), fall 
 *         back to T1 if not in any of the three regions, and use "<" instead 
 *         of "<<" for the regions
 *  T5EQ : Leung, Gammie, Noble Eq.(72), with K2(1/Tdl) ~ 2 Tdl^2
 *  T6EQ : Leung, Gammie, Noble Eq.(72)
 *  T7EQ : Vath95 Eq.(19) and Kirchoff's Law
 *  T8EQ : Chanmugam89 Eq.(1) and Kirchoff's Law
 * == Nonthermal Synchrotron ==
 *  N0EQ : Ozel00 Eq.(19) and (21)
 * == Exact integrator ==
 *  EXEQ : Wardzinski00 Eq.(3), with SN's bessel subroutine
 *
 * == About options of T0EQ ==
 * T0EQ_OPTION
 *      = T0OPT_0 : use I(x_M) on Pacholcyzk70 p.98
 *        T0OPT_1 : I(x_M) is MNY96 Eq.(29)
 *        T0OPT_2 : I(x_M) here is I'(x_M) in MNY96 Eq.(32) and Eq.(26).
 *                  It is the angle-averaged I(x_M).
 */
/* == References ==
 *  [Baring88] Baring 1988, MNRAS, 235, 51
 *  [Bekefi66] Bekefi 1966, Radiation Processes in Plasmas
 *  [Brainerd87] Brainerd & Petrosian 1987, ApJ, 320, 703
 *  [Jones79] Jones & Hardee 1979, ApJ, 228, 268
 *  [MNY96] Mahadevan, Narayan & Yi 1996, ApJ, 465, 327
 *  [Pacholczyk70] Pacholczyk 1970, Radio Astrophysics
 *  [PadI] Padmanabhan 2000, Theoretical Astrophysics, Vol. I
 *  [Petrosian81] Petrosian 1981, ApJ, 251, 727
 *  [Rybicki79] Rybicki & Lightman 1979, Radiative Processes in Astrophysics
 *  [Takahara82] Takahara & Tsuruta 1982, Prog. of Theo. Phy., 67, 485
 *  [Vath95] Vath & Chanmugam 1995, ApJS, 98, 295
 *  [Wardzinski00] Wardzinski & Zdziarski 2000, MNRAS, 314, 183
 *  [YQN03] Yuan, Quataert & Narayan 2003, ApJ, 598, 301
 */

/* -------------------------------------------------------------------------- */

//#include "nrutil.h"
#if (STAND_ALONE_EMISS_SUBROUTINE == 0)
 #include "decs.h"
 #include "global.h"
#else
 #include <stdio.h>
 #include <stdlib.h>
 #include <math.h>
 #include <float.h>
 #include <ctype.h>
 #include <math.h>
 //#include <mathimf.h>
 /* Mnemonics for rad. trans. equations EMISSION_EQ*/
 #define B0EQ   (100)          /* Bremsstrahlung:  B0EQ : PI Eq.(6.227) */
 #define C0EQ   (600)          /* Compton Emission assuming isotropy: PI Eq.(6.76) */
 #define T0EQ   (200)          /* Thermal Synch.: Pacholczyk70 Eq.(3.54), with I(x_M) set by using T0_OPTION */
 #define T1EQ   (201)          /* Thermal Synch.: Wardzinski00 Eq.(7)-(10)         */
 #define T2EQ   (202)          /* Thermal Synch.: Baring88 Eq.(10a) */
 #define T3EQ   (203)          /* Thermal Synch.: Wardzinski00 Eq.(6) */
 #define T4EQ   (204)          /* Thermal Synch.: Baring88 Eq.(10) */
 #define T5EQ   (205)          /* Thermal Synch.: Leung, Gammie, Noble Eq.(72), with K2(1/Tdl) ~ 2 Tdl^2 */
 #define T6EQ   (206)          /* Thermal Synch.: Leung, Gammie, Noble Eq.(72) */
 #define T7EQ   (207)          /* Thermal Synch.: Vath95 Eq.(19) and Kirchoff's Law */
 #define T8EQ   (208)          /* Thermal Synch.: Chanmugam89 Eq.(1) and Kirchoff's Law */
 #define N0EQ   (300)          /* Nonthermal Synch.: Ozel00 Eq.(19),(21) */
 #define T0_B0EQ     (500)     /* T0 Synch. w/ B0 Brem.                 */
 #define T1_B0EQ     (501)     /* T1 Synch. w/ B0 Brem.                 */
 #define T2_B0EQ     (502)     /* T2 Synch. w/ B0 Brem.                 */
 #define T3_B0EQ     (503)     /* T3 Synch. w/ B0 Brem.                 */
 #define T4_B0EQ     (504)     /* T4 Synch. w/ B0 Brem.                 */
 #define T5_B0EQ     (505)     /* T5 Synch. w/ B0 Brem.                 */
 #define T6_B0EQ     (506)     /* T6 Synch. w/ B0 Brem.                 */
 /* Mnemonics for options for T0EQ */
 #define T0OPT_0 (0)           /* use I(x_M) on Pacholcyzk70 p.98 */
 #define T0OPT_1 (1)           /* I(x_M) is MNY96 Eq.(29) */
 #define T0OPT_2 (2)           /* I(x_M) here is I'(x_M) in MNY96 Eq.(32) and Eq.(26). It is the angle-averaged I(x_M). */
 /* == Definition of constants in CGS units (duplicate here so that this subroutine can be used alone) == */
 #define C_1_3      (0.33333333333333333333      ) /* 1/3 */
 #define C_1_6      (0.16666666666666666666      ) /* 1/6 */
 #define C_1_12     (0.083333333333333333333     ) /* 1/12 */
 #define C_2_12     (0.16666666666666666666      ) /* 2/12 */
 #define C_19_12    (1.5833333333333333333       ) /* 19/12 */
 #define C_22_12    (1.8333333333333333333       ) /* 22/12 */
 #define C_pi       (3.1415926535897932384       ) /* pi */
 #define C_2pi      (6.2831853071795864769       ) /* 2*pi */
 #define C_4pi      (12.5663706143591729538      ) /* 4*pi */
 #define C_pi_2     (1.5707963267948966192       ) /* pi/2 */
 #define C_euler    (0.5772156649015328606065    ) /* Euler's constant */
 #define C_3_to_1_2 (1.73205080756887729352744634) /* 3^(1/2) */
 #define C_2_to_11_12 (1.887748625363386993283826) /* 2^(11/12) */
 #define C_sqrt2    (1.414213562373095048801689  ) /* sqrt(2) */
 #define C_sqrtpi   (1.77245385090551602729      ) /* sqrt(pi) */
 #define C_sqrthalfpi (1.25331413731550025121    ) /* sqrt(pi/2) */
 #define C_sqrt3pi  (3.06998012383946546542      ) /* sqrt(3*pi) */
 #define C_4pi__sqrt3_gamma1over3_2to1over3 (2.14952824153447863671) /* 4 pi / (sqrt(3) Gamma(1/3) 2^(1/3)) */
 #define C_e        (4.80320680e-10              ) /* electron charge */
 #define C_c        (2.99792458e10               ) /* speed of light */
 #define C_me       (9.1093826e-28               ) /* electron mass */
 #define C_mp       (1.67262171e-24              ) /* proton mass */
 #define C_mn       (1.67492728e-24              ) /* neutron mass */
 #define C_amu      (1.66053886e-24              ) /* atomic mass unit */
 #define C_mAtom    (1.00794 * C_amu             ) /* atomic mass unit */
 #define C_h        (6.6260693e-27               ) /* Planck constant */
 #define C_k        (1.3806505e-16               ) /* Boltzmann constant */
 #define C_G        (6.6742e-8                   ) /* Gravitational constant */
 #define C_c2       (8.9875517873681764e20       ) /* c^2 */
 #define C_c3       (2.6944002417373989539336e31 ) /* c^3 */
 #define C_c5       (2.4216061708512206534320e52 ) /* c^5 */
 #define C_e2       (2.3070795563566238161025e-19) /* e^2 */
 #define C_e3       (1.1081380213233118296897e-28) /* e^3 */
 #define C_me2      (8.2980851353182789400132e-55) /* me^2 */
 #define C_me3      (7.5590432344986989028955e-82) /* me^3 */
 #define C_me_c2    (8.1871047868450580395355e-7 ) /* me*c^2 */
 #define C_e__2pi_me_c (2.799250542245159283e6   ) /* e / (2 Pi m_e c) : this factor times B is nu_c */
 #define C_k__me_c2 (1.686372088724713698e-10    ) /* k / (me c^2) : this factor times temperature is Tdl */
 #define C_4pi_m_c__3_e (2.3815898455876043578e-7) /* 4 pi m_e c / (3 e) */
 #define C_1__me_plus_mp (5.9753838745824098154e23) /* 1 / (m_e + m_p) */
 #define C_e3__8_sqrt3_pi2_me (8.895176161867948939e-4) /* e^3 / (8 Sqrt(3) pi^2 m_e) */
 #define C_1AU      (1.49597870691e13            ) /* Astronomical Unit */
 #define C_nA       (6.0221415e23                ) /* Avogadro number */
 #define C_eV       (1.60217653e-12              ) /* electron volt in erg */
 #define C_sigma    (5.670400e-5                 ) /* Stefan-Boltzmann constant */
 #define C_alpha    (7.297352568e-3              ) /* fine structure constant */
 #define C_rydberg  (2.17987209e-11              ) /* Rydberg constant times hc in erg */
 #define C_sigmaT   (0.665245873e-24             ) /* Thomson cross section in cm^2 */
 #define C_Jy       (1.e-23                      ) /* Jansky (flux/freq. unit) in cgs */
 #define C_bohr_mag (9.2740094980e-27            ) /* Bohr Magneton (electron) */
 #define C_nucl_mag (5.0507834343e-30            ) /* Nuclear Magneton (proton) */
 #define C_pc       (3.085678e18                 ) /* parsec */
 #define C_yr       (31536000.                   ) /* No. of seconds in year */
 #define C_day      (86400.                      ) /* No. of seconds in day  */
 #define C_hr       (3600.                       ) /* No. of seconds in hour */
 #define C_ly       (9.454254955488e17           ) /* light-year (distance in cm) */
 #define C_mSUN     (1.99e33                     ) /* solar mass */
 #define C_rSun     (6.96e10                     ) /* Radius of Sun */
 #define C_LSun     (3.9e33                      ) /* Luminousity of Sun */
 #define C_TSun     (5.78e3                      ) /* Temperature of Sun's photosphere */
 #define C_MEarth   (5.976e27                    ) /* Earth's mass */
 #define C_rEarth   (6.378e8                     ) /* Earth's radius */
 #define C_dSgrA    (8.e3 * C_pc                 ) /* Distance from Earth to Sgr A*  */
#endif

//------------------------------------------------------------------------------

/* -------------------------------------------------------------------------- */

double K2_T1EQ(double Tdl) ;
void K2_T1EQ_stable(double Tdl, double *factor, double *exponent) ;
#if ( STAND_ALONE_EMISS_SUBROUTINE == 0)
 extern double exp_factor(  double f_factor, double f_exp )  ;
 #define EXP_FACTOR exp_factor
#endif

#if ( STAND_ALONE_EMISS_SUBROUTINE == 1)
 #define EXP_FACTOR exp_factor_emiss
 extern double EXP_FACTOR(  double f_factor, double f_exp )  ;
 extern int T0EQ_OPTION ;
 int T0EQ_OPTION ;
#endif

/******************************************************************************/
/*                           For common use                                   */
/******************************************************************************/
/* exp_factor_emiss():
 * -- given f_exp, and f_factor, returns with   f_factor * exp( f_exp ) ;
 * -- performs this calculation more accurately than the straightforward way;
 */

#if ( STAND_ALONE_EMISS_SUBROUTINE == 1)
double exp_factor_emiss(  double f_factor, double f_exp )
{
   double log_f, sign_f, fabs_exp, x;

   if( f_factor == 0. ) { return( 0. ); }

   fabs_exp = fabs(f_exp);

   /* If it is small, then use a 9th-order Taylor expansion of the exponent : */
   if ( fabs_exp < 1.e-3 ) {
      x = f_exp;
      return(  f_factor
               * ( 1. + ((40320.+(20160.+(6720.+(1680.+(336.+(56.+(8.+x)*x)*x)*x)*x)*x)*x)*x / 40320.) ) ) ;
   }
   /* We want to prevent overflows, calculate exponential wisely if we are close: */
   else if( fabs_exp > 690. ) {
      sign_f = ( f_factor < 0. ) ? -1. : 1. ;
      log_f = log( fabs(f_factor)  );
      if( log_f * f_exp < 0. ) {
         return( sign_f * exp( log_f + f_exp ) );
      }
      else {
         return( f_factor * exp( f_exp ) );
      }
   }
   /* Standard expression: */
   else {
      return( f_factor * exp( f_exp ) );
   }
}
#endif


/******************************************************************************/
/*                           Extra subroutines                                */
/******************************************************************************/

/* {{{--- for B0EQ ---------------------------------------------------------- */
/* velocity-averaged Gaunt factor (PadI Eq.(6.230)) */
double gffFun_B0(double nu, double Te)
{
   return(11.69 + 0.83*log(Te/1.e4) - 0.55*log(nu/1.e8)) ;
}

/*  spec. emiss. of e^- (in erg cm^-3 s^-1 ster^-1 Hz^-1) for equation B0 */
double jnu_B0(double nu, double rho, double Te)
{
   double ans ;
   /* == Variables ==
    * == (For all the EMISSION_EQ) ==
    *  nrhoe : number density of electron
    *  kT : k * Te
    *  hnu : h * nu
    */
   double nrhoe,kT,hnu ;

   kT = C_k * Te ;
   hnu = C_h * nu ;
   nrhoe = rho/(C_me + C_mp) ;

   ans = EXP_FACTOR( (6.8e-38*nrhoe*nrhoe*gffFun_B0(nu,Te)/(C_4pi * sqrt(Te)) ), 
           (-hnu/kT) );

   return(ans) ;
}
/* ------ for B0EQ -------------------------------------------------------}}} */

/* {{{--- for N0EQ ---------------------------------------------------------- */
/*  Ozel00 Eq.(6) */
/*  use the approximate equation Ozel00 Eq.(7) */
double a_N0EQ(double Tdl)
{
   return( (6+15*Tdl)/(4+5*Tdl) ) ;
}

/*  spec. emiss. of e^- (in erg cm^-3 s^-1 ster^-1 Hz^-1) for equation N0 */
/*  Ozel00 Eq.(19) */
double jnu_N0(double nu, double rho, double Te, double B, double theta)
{
   double ans ;
   /* == Variables ==
    *  nrhoe : number density of electron
    *  nu_c : cyclotron frequency
    *  Tdl : dimensionless plasma temperature
    */
   double p,eta ;
   double p4,p_m1 ;
   double nrhoe, nu_c, Tdl ;
   double Cpl ;

   nrhoe = rho / (C_me + C_mp) ;
   nu_c = C_e * B / (C_2pi*C_me*C_c) ;
   Tdl = C_k * Te / C_me_c2 ;

   p = NONTHERMAL_p ;
   eta = NONTHERMAL_eta ;

   p4 = p*0.25 ;
   p_m1 = p-1 ;
   Cpl = C_sqrtpi * 0.25 * pow(3,0.5*p) * ((p_m1)*(p-2)/(p+1))
         * gsl_sf_gamma(p4 + C_19_12)
         * gsl_sf_gamma(p4 - C_1_12)
         * gsl_sf_gamma(p4 + 1.25)
         / gsl_sf_gamma(p4 + 1.75) ;
   ans = Cpl*eta*(C_e2*nrhoe/C_c)*a_N0EQ(Tdl)*Tdl*nu_c*pow((nu/nu_c),-0.5*p_m1) ;

   return(ans) ;
}

/*  absorption coeff. of e^- (in cm^-1) for equation N0 */
/*  Ozel00 Eq.(21) */
double alphanu_N0(double nu, double rho, double Te, double B, double theta)
{
   double ans ;
   /* == Variables ==
    *  nrhoe : number density of electron
    *  nu_c : cyclotron frequency
    *  Tdl : dimensionless plasma temperature
    */
   double p,eta ;
   double p4 ;
   double nrhoe, nu_c, Tdl ;
   double Cpl ;

   nrhoe = rho / (C_me + C_mp) ;
   nu_c = C_e * B / (C_2pi*C_me*C_c) ;
   Tdl = C_k * Te / C_me_c2 ;

   p = NONTHERMAL_p ;
   eta = NONTHERMAL_eta ;

   p4 = p*0.25 ;
   Cpl = C_sqrt3pi * 0.125 * pow(3,0.5*p) * ((p-1)*(p-2)/C_me)
         * gsl_sf_gamma(p4 + C_2_12)
         * gsl_sf_gamma(p4 + C_22_12)
         * gsl_sf_gamma(p4 + 1.5)
         / gsl_sf_gamma(p4 + 2) ;
   ans = Cpl*eta*(C_e2*nrhoe/C_c)*a_N0EQ(Tdl)*Tdl*nu_c*pow((nu_c/nu),0.5*(p+3))/nu ;

   return(ans) ;
}
/* ------ for N0EQ -------------------------------------------------------}}} */

/* {{{--- for T0EQ ---------------------------------------------------------- */
/* {{{------ for T0EQ_OPTION == T0OPT_0 ------------------------------------- */
/* I(x_M) on Pacholcyzk70 p.98. Linear intrapolation on log-log plot of 
 * numbers in Appendix 2, and asymptotic on p.98 */
double I_xM_T0EQ_OPTION_0(double xM)
{
   int nData = 30 ;
   double xData[] = {1.e-6,   2.e-6,   5.e-6,   1.e-5,   2.e-5,   5.e-5,
                     1.e-4,   2.e-4,   5.e-4,   1.e-3,   2.e-3,   5.e-3,
                     1.e-2,   2.e-2,   5.e-2,   1.e-1,   2.e-1,   5.e-1,
                     1.e0,    2.e0,    5.e0,    1.e1,    2.e1,    5.e1,
                     1.e2,    2.e2,    5.e2,    1.e3,    2.e3,    5.e3} ;
/* PKL: I believe that there's a typo in the book for this line: */
//                     1.e-4,   4.e-4,   5.e-4,   1.e-3,   2.e-3,   5.e-3,
   double IData[] = {2.56e4,  1.61e4,  8.75e3,  5.51e3,  3.47e3,  1.88e3,
                     1.19e3,  7.47e2,  4.05e2,  2.54e2,  1.60e2,  8.58e1,
                     5.34e1,  3.31e1,  1.73e1,  1.04e1,  6.10e0,  2.86e0,
                     1.52e0, 7.48e-1, 2.50e-1, 9.24e-2, 2.81e-2, 3.90e-3,
                     5.85e-4,5.59e-5, 9.98e-7, 1.92e-8,1.35e-10,2.64e-14} ;
   double ans ;
   int i ;


   if(xM<xData[0]) {
      ans = 2.56 * pow(xM,(-2./3.)) ;
      return(ans) ;
   } else if(xM>=xData[nData-1]) {
      i = nData-1 ;
      ans = IData[i]
            * pow( (IData[i-1]/IData[i]) ,
                   (log(xM/xData[i])/log(xData[i-1]/xData[i]))
              ) ;
      return(ans) ;
   } else {
      for(i=1;i<nData;i++) {
         if(xM<xData[i]) {
            ans = IData[i]
                  * pow( (IData[i-1]/IData[i]) ,
                         (log(xM/xData[i])/log(xData[i-1]/xData[i]))
                    ) ;
            return(ans) ;
         }
      }
   }

   fprintf(stderr,"ERROR: can't find any intrapolated value of I(xM)\n") ;
   fprintf(stderr,"       in function I_xM_T0(double xM)\n") ;
   exit(1) ;
   return(0) ;
}

/* I(x_M) on Pacholcyzk70 p.98. Linear intrapolation on log-log plot of 
 * numbers in Appendix 2, and asymptotic on p.98
 * This route returns the exponent and factor parts separately so one can use 
 * exp_factor() */
void I_xM_T0EQ_OPTION_0_stable(double xM, double *factor, double *exponent)
{
   *factor = I_xM_T0EQ_OPTION_0(xM) ;
   *exponent = 0. ;
}
/* --------- for T0EQ_OPTION == T0OPT_0 ----------------------------------}}} */
/* {{{------ for T0EQ_OPTION == T0OPT_1 ------------------------------------- */
/* I(x_M) here is MNY96 Eq.(29). */
double I_xM_T0EQ_OPTION_1(double xM)
{
   double xM13,xM23 ;

   xM13 = pow(xM,-1./3.) ;
   return( EXP_FACTOR( 
            (2.5651 * (1. + xM13*(1.92 + 0.9977*xM13))),
            (-1.8899/xM13) 
            ) );

}

/* I(x_M) here is MNY96 Eq.(29).
 * This route returns the exponent and factor parts separately so one can use 
 * exp_factor() */
void I_xM_T0EQ_OPTION_1_stable(double xM, double *factor, double *exponent)
{
   double xM13 ;

   xM13 = pow(xM,-1./3.) ;
   *factor = 2.5651 * (1. + xM13*(1.92 + 0.9977*xM13)) ;
   *exponent = -1.8899/xM13 ;

}
/* --------- for T0EQ_OPTION == T0OPT_1 ----------------------------------}}} */
/* {{{------ for T0EQ_OPTION == T0OPT_2 ------------------------------------- */
/* I(x_M) here is I'(x_M) in MNY96 Eq.(32) and Eq.(26). It is the 
 * angle-averaged I(x_M). */
double I_xM_T0EQ_OPTION_2(double xM)
{
   double xM12,xM13,xM14,xM16 ;

   xM14 = pow(xM,-0.25) ;
   xM12 = xM14*xM14 ;
   xM16 = pow(xM,-(1./6.)) ;
   xM13 = xM16*xM16 ;
   return(  EXP_FACTOR( 
             (4.0505*xM16 * (1. + 0.4*xM14 + 0.5316*xM12) ), 
             (-1.8899/xM13) 
             )  );

}

/* I(x_M) here is I'(x_M) in MNY96 Eq.(32) and Eq.(26). It is the 
 * angle-averaged I(x_M).
 * This route returns the exponent and factor parts separately so one can use 
 * exp_factor() */
void I_xM_T0EQ_OPTION_2_stable(double xM, double *factor, double *exponent)
{
   double xM12,xM13,xM14,xM16 ;

   xM14 = pow(xM,-0.25) ;
   xM12 = xM14*xM14 ;
   xM16 = pow(xM,-(1./6.)) ;
   xM13 = xM16*xM16 ;
   *factor = 4.0505*xM16 * (1. + 0.4*xM14 + 0.5316*xM12) ;
   *exponent = -1.8899/xM13 ;

}

/* --------- for T0EQ_OPTION == T0OPT_2 ----------------------------------}}} */

/*  spec. emiss. of e^- (in erg cm^-3 s^-1 ster^-1 Hz^-1) for equation T0 */
double jnu_T0(double nu, double rho, double Te, double B, double theta)
{
   double ans ;
   double nuM,xM ;
   /* == Variables ==
    *  nrhoe : number density of electron
    *  nu_c : cyclotron frequency
    *  Tdl : dimensionless plasma temperature
    *  K2 : Bessel function K_2
    *  K2_factor, K2_exp : K_2 = K2_factor * Exp(K2_exp)
    */
   double nrhoe, nu_c, Tdl ;
   double K2_factor, K2_exp, I_xM_factor, I_xM_exp ;

   nrhoe = rho / (C_me + C_mp) ;
   nu_c = C_e * B / (C_2pi*C_me*C_c) ;
   Tdl = C_k * Te / C_me_c2 ;
   /* x_M = nu / nu_M, where x_M is same as the definition in MNY96 Eq.(12) */
   nuM = 1.5 * nu_c * Tdl * Tdl ;

   if(T0EQ_OPTION == T0OPT_0) {
      /* x_M as defined on Pacholczyk 70 page 98 */
      xM = nu / (nuM * sin(theta)) ;
      I_xM_T0EQ_OPTION_0_stable(xM, &I_xM_factor, &I_xM_exp) ;
   } else if(T0EQ_OPTION == T0OPT_1) {
      /* x_M here is x_M / sin(theta) in MNY96 Eq.(11) */
      xM = nu / (nuM * sin(theta)) ;
      I_xM_T0EQ_OPTION_1_stable(xM, &I_xM_factor, &I_xM_exp) ;
   } else if(T0EQ_OPTION == T0OPT_2) {
      /* x_M here is x_M in MNY96 Eq.(26) */
      xM = nu / nuM ;
      I_xM_T0EQ_OPTION_2_stable(xM, &I_xM_factor, &I_xM_exp) ;
   } else {
      fprintf(stderr,"ERROR: undefined value of option for T0EQ!") ;
      exit(1) ;
   }
   K2_T1EQ_stable(Tdl, &K2_factor, &K2_exp) ;
   ans = EXP_FACTOR(
           ( C_4pi*nrhoe*C_e2*nu*I_xM_factor/(C_3_to_1_2*C_c*K2_factor) ),
           ( I_xM_exp - K2_exp )
         ) ;

   return(ans) ;
}

#if (STAND_ALONE_EMISS_SUBROUTINE ==1)
/* wrappers: */
double jnu_T0_0(double nu, double rho, double Te, double B, double theta)
{
   T0EQ_OPTION = T0OPT_0 ;
   return(jnu_T0(nu,rho,Te,B,theta)) ;
}
double jnu_T0_1(double nu, double rho, double Te, double B, double theta)
{
   T0EQ_OPTION = T0OPT_1 ;
   return(jnu_T0(nu,rho,Te,B,theta)) ;
}
double jnu_T0_2(double nu, double rho, double Te, double B, double theta)
{
   T0EQ_OPTION = T0OPT_2 ;
   return(jnu_T0(nu,rho,Te,B,theta)) ;
}
#endif
/* ------ for T0EQ -------------------------------------------------------}}} */

/* {{{--- for T1EQ ---------------------------------------------------------- */
/* Wardzinski00 Eq.(4) */
double K2_T1EQ(double Tdl)
{
  /* == input variables ==
   *  Tdl : dimensionless plasma temperature
   */
  double tmp,Tdl2,ans ;

  if(Tdl<=0.65) {
    ans = EXP_FACTOR(  
           (sqrt(C_pi_2*Tdl) * (1. + Tdl*((15./8.) + Tdl*((105./128.) - 0.203*Tdl)))),
           (-1./Tdl) 
           );
  } else {
    tmp = log(2.*Tdl) ;
    Tdl2 = 8.*Tdl*Tdl ;
    ans = 0.25*Tdl2 - 0.5 + ((tmp+0.75-C_euler) + (tmp+0.95)/(1.5*Tdl2))/Tdl2;
  }

  if( !finite(ans) ) { 
    fprintf(stderr,"K2_T1EQ(): ans,Tdl  = %28.18e %28.18e \n", ans,Tdl);
    fflush(stderr);
  }   

  return(ans);
}

/* Wardzinski00 Eq.(4) that returns the exponent and factor parts separately so one can use exp_factor() */
void K2_T1EQ_stable(double Tdl, double *factor, double *exponent)
{
  /* == input variables ==
   *  Tdl : dimensionless plasma temperature
   */
  double tmp,Tdl2;

  if(Tdl<=0.65) {
    *factor   = (sqrt(C_pi_2*Tdl) * (1. + Tdl*((15./8.) + Tdl*((105./128.) - 0.203*Tdl)))) ;
    *exponent = (-1./Tdl) ;
  } else {
    tmp = log(2.*Tdl) ;
    Tdl2 = 8.*Tdl*Tdl ;
    *factor = 0.25*Tdl2 - 0.5 + ((tmp+0.75-C_euler) + (tmp+0.95)/(1.5*Tdl2))/Tdl2;
    *exponent = 0.;
  }

  if( !finite(*exponent) || !finite(*factor) ) { 
    fprintf(stderr,"K2_T1EQ_stable(): factor,exponent,Tdl  = %28.18e %28.18e %28.18e \n", *factor, *exponent,Tdl);
    fflush(stderr);
  }   

  return;
}

/* Wardzinski00 Eq.(3) */
double ne_T1EQ(double Tdl, double nrhoe, double gammaL)
{
   /* == input variables ==
    *  Tdl : dimensionless plasma temperature
    *  nrhoe : electron number density
    *  gammaL : Lorentz factor of electron (thermal motion, not the bulk motion)
    */
  double ans, K2_factor, K2_exp;

  K2_T1EQ_stable(Tdl, &K2_factor, &K2_exp);
  ans =  EXP_FACTOR( 
          (nrhoe*gammaL*sqrt((gammaL-1.)*(gammaL+1.))/(Tdl*K2_factor) ),
          (-gammaL/Tdl - K2_exp)
          );

  if( !finite(ans) ) { 
    fprintf(stderr,"ne_T1EQ(): ans,Tdl,nrhoe,gammaL,K2 = %28.18e %28.18e %28.18e %28.18e %28.18e  %28.18e \n",
       ans,Tdl,nrhoe,gammaL,K2_factor, K2_exp);
    fflush(stderr);
  }   

  return(ans);
}

/* Wardzinski00 Eq.(8) */
double Z_T1EQ(double gammaL, double sinangle, double nuratio)
{
   /* == input variables ==
    *  gammaL : Lorentz factor of electron (thermal motion, not the bulk motion)
    *  sinangle : sin(angle), here angle is "the angle between B and the 
    *     direction towards the observer"
    *  nuratio : nu/nu_c, where nu_c is cyclotron frequency
    */
   double t,tmp,tmpSqrt,ans ;

   t = sqrt((gammaL-1.)*(gammaL+1.))*sinangle ;
   tmp = 1.+t*t ;
   tmpSqrt = sqrt(tmp) ;
   ans = EXP_FACTOR( 1. ,  ( 2.*nuratio*( tmp*log(t/(1.+tmpSqrt)) + tmpSqrt ) / gammaL ) );

   if( !finite(ans) ) { 
     fprintf(stderr,"Z_T1EQ(): ans,gammaL,sin,nurat = %28.18e %28.18e %28.18e %28.18e \n",
        ans,gammaL,sinangle,nuratio);
     fflush(stderr);
   }   

   return(ans);
}

/* Wardzinski00 Eq.(9) */
double Chi_T1EQ(double Tdl, double gammaL)
{
   /* == input variables ==
    *  Tdl : dimensionless plasma temperature
    *  gammaL : Lorentz factor of electron (thermal motion, not the bulk motion)
    */
   double gammaL2,ans ;

   if(Tdl<=0.08) {
     gammaL2 = gammaL * sqrt(3.);
     ans = sqrt(2.*Tdl*(gammaL-1.)*(gammaL+1.)/(gammaL*(gammaL2-1.)*(gammaL2+1.)));
   } else {
     ans = sqrt(Tdl/(1.5*gammaL));
   }

   if( !finite(ans) ) { 
     fprintf(stderr,"Chi_T1EQ(): ans,Tdl,gammaL = %28.18e %28.18e %28.18e \n",ans,Tdl,gammaL);
     fflush(stderr);
   }   

   return(ans);
}

/* Wardzinski00 Eq.(10) */
double gamma0_T1EQ(double Tdl, double sinangle, double nuratio)
{
   /* == input variables ==
    *  Tdl : dimensionless plasma temperature
    *  sinangle : sin(angle), here angle is "the angle between B and the 
    *     direction towards the observer"
    *  nuratio : nu/nu_c, where nu_c is cyclotron frequency
    */
   double tmp,ans;

   tmp = nuratio*Tdl ;

   if(Tdl<=0.08) {
     ans = sqrt(1. + 2.*tmp / pow((1.+4.5*tmp*sinangle*sinangle),1./.3)) ;
   } else {
     ans = sqrt(1. + pow( (tmp/(0.75*sinangle)), (2./3.) ) ) ;
   }

   if( !finite(ans) ) { 
     fprintf(stderr,"gamma0_T1EQ(): ans,Tdl,sin,nurat = %28.18e %28.18e %28.18e %28.18e \n",
        ans,Tdl,sinangle,nuratio);
     fflush(stderr);
   }   

   return(ans) ;
}

/*  spec. emiss. of e^- (in erg cm^-3 s^-1 ster^-1 Hz^-1) for equation T1 */
double jnu_T1(double nu, double rho, double Te, double B, double theta)
{
   double ans ;
   /* == Variables ==
    *  gamma0 : Lorentz factor in Wardzinski00 Eq.(10)
    *  kT : k * Te
    *  nrhoe : number density of electron
    *  nu_c : cyclotron frequency
    *  nuratio : nu/nu_c
    *  sinangle : sin(theta)
    *  Tdl : dimensionless plasma temperature
    */
   double gamma0,kT,nrhoe,nu_c,nuratio,Tdl ;
   double sinangle,cosangle,cotangle, inv_gammasq ;
   double gammasq_m1, sin_sq, tvar, svar, t_sq,s_sq;
   double Eexp,kappa,avar,bvar,phivar,sqrt_term,tmp, slog_term;
   double t1;

   kT = C_k * Te ;
   nu_c = C_e*B/(C_2pi*C_me*C_c) ;
   nrhoe = rho/(C_me + C_mp) ;
   nuratio = nu/nu_c ;
   Tdl = kT/C_me_c2 ;

   tmp = nuratio*Tdl ;
   //sincos( theta, &sinangle, &cosangle );
   sinangle = sin(theta);
   cosangle = cos(theta);
   if(sinangle==0.) {
     ans = 0. ;
   } else {

     cotangle = cosangle/sinangle ;
     sin_sq = sinangle*sinangle;

     if(Tdl<=0.08) {
       gammasq_m1 = 2.*tmp / pow((1.+4.5*tmp*sin_sq), (1./.3)) ;
       sqrt_term = gammasq_m1 / sqrt( 2.*Tdl*(3.*gammasq_m1 + 2.) );
     } else {
       gammasq_m1 = pow( (tmp/(0.75*sinangle)), (2./3.) )  ;
       sqrt_term = sqrt( gammasq_m1  / (6.*Tdl) );
     }
     
     gamma0 = sqrt( gammasq_m1 + 1. );
     inv_gammasq = 1./ ( gammasq_m1 + 1. ) ; 

     t_sq = gammasq_m1 * sin_sq;
     tvar = sqrt( t_sq );
     s_sq = t_sq + 1.;
     svar = sqrt( s_sq ) ; 

     if( svar > 1.e1 ) { 
       t1 = 1./s_sq;
       slog_term = ( -(15015.0+(9009.0+(6435.0+(5005.0+(4095.0+(3465.0+3003.0*t1)*t1)*
                   t1)*t1)*t1)*t1)/(45045.0*svar));
     } 
     else { 
       slog_term = svar * ( 1. + 0.5*svar*log((svar-1.)/(svar+1.)) ) ; 
     }

     if( Tdl <= 0.65 ) { 
       Eexp = -gammasq_m1/((gamma0+1.)*Tdl) + 2.*nuratio*slog_term/gamma0 ;
       kappa = sqrt( 0.5*C_pi*Tdl ) * ( 1. + Tdl * ( (15./8.) + Tdl * ( (105./128.) - 0.203*Tdl ) ) ) ;
     }
     else { 
       Eexp = -gamma0/Tdl                   + 2.*nuratio*slog_term/gamma0 ;
       phivar = 2.*Tdl;
       bvar = log(phivar);
       phivar *= phivar;
       avar = bvar + 0.75 - C_euler;
       bvar += 0.95;
       kappa = ( bvar + 3.*phivar * ( avar + phivar * ( phivar - 1. ) ) ) / ( 6. * phivar*phivar ) ;;
     }

     ans = EXP_FACTOR( ( (C_pi*C_e2/(C_c*kappa)) * nrhoe 
          * ( 1. + 2.*cotangle*cotangle*inv_gammasq ) 
          * sqrt(nu*nu_c*svar) 
          * sqrt_term ) , 
             Eexp );

     if( 0 ||  !finite(ans) ) { 
       fprintf(stderr,"jnu_T1()2: k,a,b,Eexp,Tdl,gam  = %28.18e %28.18e %28.18e %28.18e %28.18e %28.18e \n",
          kappa,avar,bvar,Eexp,Tdl,gamma0);
       fprintf(stderr,"jnu_T1()2: phi,nuc,s,t,cot,ne = %28.18e %28.18e %28.18e %28.18e %28.18e %28.18e  \n",
          phivar,nu_c, svar, tvar,cotangle,nrhoe );
       fflush(stderr);
     }   

   }

   if( 0 || !finite(ans) ) { 
     fprintf(stderr,"jnu_T1(): ans,nu,rho,Te,B,theta = %28.18e %28.18e %28.18e %28.18e %28.18e %28.18e \n",
        ans,nu,rho,Te,B,theta);
     fflush(stderr);
   }   


   return(ans) ;
}
/* ------ for T1EQ -------------------------------------------------------}}} */

/* {{{--- for T2EQ ---------------------------------------------------------- */
/*  T2EQ : Baring88 Eq.(10a), see also Brainerd87 Eqs.(39), (40) and (45) */
/*  spec. emiss. of e^- (in erg cm^-3 s^-1 ster^-1 Hz^-1) for equation T2 */
double jnu_T2(double nu, double rho, double Te, double B, double theta)
{
   double ans ;
   /* == Variables ==
    *  kT : k * Te
    *  nrhoe : number density of electron
    *  nu_c : cyclotron frequency
    *  nuratio : nu/nu_c
    *  sinangle : sin(theta)
    *  Tdl : dimensionless plasma temperature
    *  omega : photon energy scaled by electron rest mass energy
    *          = hbar \omega / (m_e c^2) = h \nu / (m_e c^2)
    *  B_c : critical field strength in Brainerd87 Eq.(7)
    *        = (m_e^2 c^3) / (hbar e) ~ 10^13 G
    *  B_perp : component (perpendicular to emission angle) of scaled field,
    *           Brainerd87 Eq.(7)
    */
   double kT,nrhoe,nu_c,nuratio,Tdl,Tdl3 ;
   double omega,B_c,B_perp ;

   kT = C_k * Te ;
   nu_c = C_e*B/(C_2pi*C_me*C_c) ;
   nrhoe = rho/(C_me + C_mp) ;
   nuratio = nu/nu_c ;
   Tdl = kT/C_me_c2 ;
   Tdl3 = Tdl*Tdl*Tdl ;

   omega = C_h * nu / C_me_c2 ;
   B_c = (C_me2 * C_c3 * C_2pi) / (C_h * C_e) ;
   B_perp = B * sin(theta) / B_c ;

   ans = (8./3.) * pow((B_perp/9.),(2./3.))
         * pow((omega*pow(Tdl,7)),(1./3.)) ;

   ans *= C_e2 * C_me_c2 * C_me_c2 * nrhoe * C_2pi * C_2pi
          / (C_h * C_h * C_c * 2. * Tdl3) ;

   ans *= C_h / (C_4pi * C_me_c2) ;

   return(ans) ;
}
/* ------ for T2EQ -------------------------------------------------------}}} */

/* {{{--- for T3EQ ---------------------------------------------------------- */
/*  T3EQ : Wardzinski00 Eq.(6) */
/*  spec. emiss. of e^- (in erg cm^-3 s^-1 ster^-1 Hz^-1) for equation T3 */
double jnu_T3(double nu, double rho, double Te, double B, double theta)
{
   double ans ;
   /* == Variables ==
    *  nrhoe : number density of electron
    *  nu_c : cyclotron frequency
    *  sinangle : sin(theta)
    *  Tdl : dimensionless plasma temperature
    *  v : nu/(nu_c Tdl^2), see the line below Wardzinski00 Eq.(6)
    */
   double nrhoe,nu_c,Tdl,v, K2_factor, K2_exp ;
   double sinangle ;

   nrhoe = rho/(C_me + C_mp) ;
   nu_c = C_e*B/(C_2pi*C_me*C_c) ;
   Tdl = C_k*Te/C_me_c2 ;
   v = nu/(nu_c*Tdl*Tdl) ;

   sinangle = sin(theta) ;

   if(sinangle==0.) {
      ans = 0. ;
   } else {
      K2_T1EQ_stable(Tdl, &K2_factor, &K2_exp ); 
      ans = EXP_FACTOR( sqrt(2.)*C_pi*C_e2*nrhoe*nu/(3.*C_c*K2_factor) , 
            ( -pow(((4.5*v)/sinangle),(1./3.)) - K2_exp ) ) ;
   }

   return(ans) ;
}

/* ------ for T3EQ -------------------------------------------------------}}} */

/* {{{--- for T4EQ ---------------------------------------------------------- */
/*  T4EQ : Baring88 Eq.(10), see also Brainerd87 Eqs.(39), (40) and (45)
 *         Fall back to T1 is not in any of the three regions */
/*  spec. emiss. of e^- (in erg cm^-3 s^-1 ster^-1 Hz^-1) for equation T4 */
double jnu_T4(double nu, double rho, double Te, double B, double theta)
{
   double ans ;
   /* == Variables ==
    *  kT : k * Te
    *  nrhoe : number density of electron
    *  nu_c : cyclotron frequency
    *  nuratio : nu/nu_c
    *  sinangle : sin(theta)
    *  Tdl : dimensionless plasma temperature
    *  omega : photon energy scaled by electron rest mass energy
    *          = hbar \omega / (m_e c^2) = h \nu / (m_e c^2)
    *  B_c : critical field strength in Brainerd87 Eq.(7)
    *        = (m_e^2 c^3) / (hbar e) ~ 10^13 G
    *  B_perp : component (perpendicular to emission angle) of scaled field,
    *           Brainerd87 Eq.(7)
    */
   double kT,nrhoe,nu_c,nuratio,Tdl,Tdl2,Tdl3 ;
   double omega,B_c,B_perp ;
   double Tdl_B_sq,omega_Tdl,TdlB ;
   //double SMALL_FAC ;
   double SMALL_FAC=0.1 ;

   kT = C_k * Te ;
   nu_c = C_e*B/(C_2pi*C_me*C_c) ;
   nrhoe = rho/(C_me + C_mp) ;
   nuratio = nu/nu_c ;
   Tdl = kT/C_me_c2 ;
   Tdl2 = Tdl*Tdl ;
   Tdl3 = Tdl*Tdl2 ;

   omega = C_h * nu / C_me_c2 ;
   B_c = (C_me2 * C_c3 * C_2pi) / (C_h * C_e) ;
   B_perp = B * sin(theta) / B_c ;

   Tdl_B_sq = sqrt(Tdl/B_perp) ;
   omega_Tdl = omega/Tdl ;

   if(omega_Tdl < 1.) {
      /* Baring 88 Eq.(10a) */
      ans = (8./3.) * pow((B_perp/9.),(2./3.))
            * pow((omega*pow(Tdl,7)),(1./3.)) ;
   } else if ((Tdl2*B_perp < omega) && (omega < Tdl_B_sq)) {
      /* Baring 88 Eq.(10b) */
      ans = EXP_FACTOR( (Tdl*sqrt(2.)/3.) * omega , 
            ( -0.5*omega_Tdl - pow((9.*omega/(2.*Tdl2*B_perp)),(1./3.)) ) ) ;
   } else if (Tdl_B_sq <= SMALL_FAC*omega) {
      /* Baring 88 Eq.(10c) */
      TdlB = Tdl*B_perp ;
      ans = EXP_FACTOR( 0.5*omega*omega*sqrt(TdlB) , 
            ( -omega_Tdl
                 - sqrt(1./(0.375*TdlB))
                 + 1./(1.5*omega*B_perp)) ) ;
   } else {
      return(jnu_T1(nu,rho,Te,B,theta)) ;
   }

   ans *= C_e2 * C_me_c2 * C_me_c2 * nrhoe * C_2pi * C_2pi
          / (C_h * C_h * C_c * 2. * Tdl3) ;

   ans *= C_h / (C_4pi * C_me_c2) ;

   return(ans) ;
}
/* ------ for T4EQ -------------------------------------------------------}}} */

/* {{{--- for T5EQ ---------------------------------------------------------- */
/*  T5EQ : Leung, Gammie, Noble Eq.(72), with K2 ~ 2 Tdl^2 */
/*  spec. emiss. of e^- (in erg cm^-3 s^-1 ster^-1 Hz^-1) for equation T5 */
double jnu_T5(double nu, double rho, double Te, double B, double theta)
{
   double ans ;
   /* == Variables ==
    *  kT : k * Te
    *  nrhoe : number density of electron
    *  nu_s : cyclotron frequency * (2/9) * Tdl^2 sin(theta)
    *  Tdl : dimensionless plasma temperature
    *  x : nu/nu_s
    */
   double kT,nrhoe,nu_s,Tdl,x ;

   kT = C_k * Te ;
   Tdl = kT/C_me_c2 ;
   nu_s = C_e*B*Tdl*Tdl*sin(theta)/(4.5*C_2pi*C_me*C_c) ;
   x = nu/nu_s ;
   nrhoe = rho/(C_me + C_mp) ;

   ans = EXP_FACTOR( ( (C_pi*C_e2*nrhoe*nu_s)/(3*C_sqrt2*C_c*Tdl*Tdl)
                       * pow( (sqrt(x) + C_2_to_11_12*pow(x,C_1_6)) , 2. ) ),
                     ( -pow(x,C_1_3) )
                   ) ;

   return(ans) ;
}
/* ------ for T5EQ -------------------------------------------------------}}} */

/* {{{--- for T6EQ ---------------------------------------------------------- */
/*  T6EQ : Leung, Gammie, Noble Eq.(72) */
/*  spec. emiss. of e^- (in erg cm^-3 s^-1 ster^-1 Hz^-1) for equation T6 */
double jnu_T6(double nu, double rho, double Te, double B, double theta)
{
   double ans ;
   /* == Variables ==
    *  kT : k * Te
    *  nrhoe : number density of electron
    *  nu_s : cyclotron frequency * (2/9) * Tdl^2 sin(theta)
    *  Tdl : dimensionless plasma temperature
    *  x : nu/nu_s
    *
    *  K2_factor, K2_exp : K_2 = K2_factor * Exp(K2_exp)
    */
   double kT,nrhoe,nu_s,Tdl,x ;
   double K2_factor, K2_exp ;

   kT = C_k * Te ;
   Tdl = kT/C_me_c2 ;
   nu_s = C_e*B*Tdl*Tdl*sin(theta)/(4.5*C_2pi*C_me*C_c) ;
   x = nu/nu_s ;
   nrhoe = rho/(C_me + C_mp) ;

   K2_T1EQ_stable(Tdl, &K2_factor, &K2_exp ); 
   ans = EXP_FACTOR( ( (C_sqrt2*C_pi*C_e2*nrhoe*nu_s)/(3*C_c*K2_factor)
                       * pow( (sqrt(x) + C_2_to_11_12*pow(x,C_1_6)) , 2. ) ),
                     ( -pow(x,C_1_3) - K2_exp )
                   ) ;

   return(ans) ;
}
/* ------ for T6EQ -------------------------------------------------------}}} */

/* {{{--- for T7EQ ---------------------------------------------------------- */
/*  T7EQ : Vath95 Eq.(19) and Kirchoff's Law */
/*  Return spec. emiss. of e^- (in erg cm^-3 s^-1 ster^-1 Hz^-1)
 *   and absorption coeff. of e^- (in erg cm^-2) for equation T7 */
void return_jnuOX_alphanuOX_T7(double *jnuO, double *jnuX,
                               double *alphaO, double *alphaX,
                               double nu, double rho, double Te,
                               double B, double theta)
{
   double kT,Tdl,nrhoe,nu_c,nu_p,nuratio,sintheta,sin2,costheta,cos2,abscos ;
   double x,gam0,g2,beta0,b2,n_refrac,n2,a ;
   double n2_b2,one_minus_n2_b2,one_minus_n2_b2_c2 ;
   double s0,betaprime,xi0,xi2,aPlus,aMinus,sc,c2Plus,c2Minus,Z,Bnu ;
   double tmp1,tmp2,tmp3,tmp4 ;
   /* == Variables ==
    *  costheta : cos(theta)
    *  kT : k * Te
    *  n_refrac : refractive index
    *  nu_c : cyclotron frequency
    *  nu_p : plasma frequency
    *  nuratio : nu/nu_c
    *  sintheta : sin(theta)
    *  Tdl : dimensionless plasma temperature
    */

   if(theta==0.) {
      *jnuO = *jnuX = *alphaO = *alphaX = 0. ;
      return ;
   }

   kT = C_k*Te ;
   Tdl = kT/C_me_c2 ;
   nrhoe = rho/(C_me + C_mp) ;
   nu_c = C_e*B/(C_2pi*C_me*C_c) ;
   nu_p = C_e*sqrt(nrhoe/(C_me*C_pi)) ;
   nuratio = nu/nu_c ;
   sintheta = sin(theta) ;
   sin2 = sintheta*sintheta ;
   costheta = cos(theta) ;
   cos2 = costheta*costheta ;
   abscos = fabs(costheta) ;

   /* Vath95 Eq.(20c) */
   x = nuratio*Tdl*sin2;
   /* Vath95 Eq.(20a) */
   g2 = 1.+2.*nuratio*Tdl*pow(1.+4.5*x,-0.33333333333333333333) ;
   gam0 = sqrt(g2) ;
   /* Vath95 Eq.(20b) */
   b2 = 1.-1./g2 ;
   beta0 = sqrt(b2) ;

   /* Vath95 Eq.(21a) */
   n_refrac = 1. - nu_p*nu_p / (nu*nu) ;
   n2 = n_refrac*n_refrac ;

   /* Vath95 Eq.(22a) */
   a = sin2 / (2.*nuratio*abscos) ;
   /* Vath95 Eq.(22b) */
   n2_b2 = n2*b2 ;
   one_minus_n2_b2 = 1. - n2_b2 ;
   one_minus_n2_b2_c2 = 1. - n2_b2*cos2 ;
   s0 = gam0*one_minus_n2_b2_c2*nuratio ;
   /* Vath95 Eq.(22c) */
   betaprime = n_refrac*beta0*sintheta/sqrt(one_minus_n2_b2_c2) ;

   /* Vath95 Eq.(21b) */
   xi2 = 1./(1.-betaprime*betaprime) ;
   xi0 = sqrt(xi2) ;
   /* Vath95 Eq.(21c) */
   aPlus = -(a + sqrt(1.+a*a))*costheta/abscos ;
   aMinus = -1./aPlus ;

   /* Vath95 Eq.(23a) */
   sc = 1.5*xi2*xi0 ;
   /* Vath95 Eq.(23b) */
   c2Plus = aPlus*costheta*one_minus_n2_b2 ;
   c2Minus = aMinus*costheta*one_minus_n2_b2 ;
   /* Vath95 Eq.(23c) */
   Z = betaprime*exp(1./xi0)/(1.+1./xi0) ;

   /* Vath95 Eq.(23c) */
   tmp1 = 2.67e-9*(nrhoe*(1.-1.875*Tdl)/(B*Tdl*Tdl*n2*sin2*sintheta))
          * sqrt(g2*gam0*(g2-1.)/(s0*s0*s0*x))*xi2*(xi2-1.)
          * one_minus_n2_b2_c2
          * pow(1. + 4.5297*sc/s0 , 0.16666666666666666666)
          * pow(Z,2.*s0)
          * exp((1.-gam0)/Tdl) ;
   tmp2 = pow(1. + 0.85*sc/s0 , -0.33333333333333333333) ;
   tmp3 = sqrt(one_minus_n2_b2*one_minus_n2_b2_c2) ;
   tmp4 = n2_b2*xi0*sin2*sin2/(2.*(s0+sc)) ;
   *alphaO = (tmp1/(1.+aPlus*aPlus))
             * ( pow(c2Plus*tmp2 + tmp3,2.) + tmp4*aPlus*aPlus ) ;
   *alphaX = (tmp1/(1.+aMinus*aMinus))
             * ( pow(c2Minus*tmp2 + tmp3,2.) + tmp4*aMinus*aMinus ) ;

   /* using Kirchoff's Law to find j_nu */
   Bnu = 2.*C_h*nu*nu*nu / ( C_c2 * (exp(C_h*nu/kT)-1.)) ;
   *jnuO = *alphaO * (Bnu*0.5) ; /* 2 in Bnu is for two polarizations */
   *jnuX = *alphaX * (Bnu*0.5) ;
}
/* ------ for T7EQ -------------------------------------------------------}}} */

/* {{{--- for T8EQ ---------------------------------------------------------- */
/*  T8EQ : Chanmugam89 Eq.(1) and Kirchoff's Law */
/*  Return spec. emiss. of e^- (in erg cm^-3 s^-1 ster^-1 Hz^-1)
 *   and absorption coeff. of e^- (in erg cm^-2) for equation T8 */
/* can't sum to infinity, so sum from nMin to nMin + T8EQ_DELTA_N - 1 : */
#define T8EQ_DELTA_N (10)
void return_jnuOX_alphanuOX_T8(double *jnuO, double *jnuX,
                               double *alphaO, double *alphaX,
                               double nu, double rho, double Te, double B)
{
   double kT,Tdl,nrhoe,nu_c,nu_p,nuratio,Bnu ;
   double tmp1,tmp2,tmp3,tmp4,sumO,sumX ;
   int n,nMin ;
   /* == Variables ==
    *  kT : k * Te
    *  n_refrac : refractive index
    *  nu_c : cyclotron frequency
    *  nu_p : plasma frequency
    *  nuratio : nu/nu_c
    *  Tdl : dimensionless plasma temperature
    */

   kT = C_k*Te ;
   Tdl = kT/C_me_c2 ;
   nrhoe = rho/(C_me + C_mp) ;
   nu_c = C_e*B/(C_2pi*C_me*C_c) ;
   nu_p = C_e*sqrt(nrhoe/(C_me*C_pi)) ;
   nuratio = nu/nu_c ;

   tmp1 = (C_4pi*nu_p*nu_p/(nu_c*C_c))*exp(2.*nuratio)*pow(2.*Tdl*nuratio,-2.5) ;
   sumO = sumX = 0. ;
   nMin = floor(nuratio)+1 ;
   for(n=nMin;n<nMin+T8EQ_DELTA_N;n++) {
     tmp2 = n/nuratio ;
     tmp3 = tmp2*tmp2 - 1. ;
     tmp4 = exp((1.-tmp2)/Tdl) * sqrt(tmp3)
            * pow((tmp2-1.)/(tmp2+1.),n) ;
     sumO += tmp4*tmp3/(2.*nuratio) ;
     sumX += tmp4 ;
   }
   *alphaO = tmp1*sumO ;
   *alphaX = tmp1*sumX ;

   /* using Kirchoff's Law to find j_nu */
   Bnu = 2.*C_h*nu*nu*nu / ( C_c2 * (exp(C_h*nu/kT)-1.)) ;
   *jnuO = *alphaO * (Bnu*0.5) ; /* 2 in Bnu is for two polarizations */
   *jnuX = *alphaX * (Bnu*0.5) ;
}
#undef T8EQ_DELTA_N
/* ------ for T8EQ -------------------------------------------------------}}} */

/* {{{--- for T9EQ ---------------------------------------------------------- */
/* T9EQ : Multiple n_e with the single particle emissivity at gamma=Theta.
 *        The single particle emissivity is obtained from
 *        Melrose 1971, ApSS, 12, 172, Eq.(37) with Taylor expansion of the
 *        Airy function and integral. */
/*  Return spec. emiss. of e^- (in erg cm^-3 s^-1 ster^-1 Hz^-1)
 *   and absorption coeff. of e^- (in erg cm^-2) for equation T7 */
/* In Melrose 1971,
 * d(j_{\omega,I})/dE == epsilon_1 = epsilon^{11} + epsilon^{22}
 * d(j_{\omega,Q})/dE == epsilon_2 = epsilon^{11} - epsilon^{22}
 * d(j_{\omega,U})/dE == epsilon_3 = epsilon^{12} + epsilon^{21} = 0
 * d(j_{\omega,V})/dE == epsilon_4 = i (epsilon^{12} - epsilon^{21})
 * 
 * Then use Eq.(37) for epsilon^{ij}, and find the Taylor expansion of
 *  the Airy function and integral.
 * 
 * d(j_{\nu,i})/d\gamma ~ m_e c^2 d(j_{\nu,i})/dE
 *                      = 2 \pi m_e c^2 d(j_{\omega,i})/dE ,
 * with i=I,Q,U,V.
 */
void return_jnuStokes_alphanuStokes_T9(
                   double *jnuI, double *jnuQ, double *jnuV,
                   double *alphaI, double *alphaQ, double *alphaV,
                   double nu, double rho, double Te, double B, double theta)
{
   double kT,Tdl,nrhoe,nu_c,nu_p,nuratio,sintheta,Bnu ;
   double gam,beta2,n_ref,xi,r,coeff ;
   double rInv,r2Inv,expMinusR,sqrtR,sqrtRInv,r2,r2_3,r4_3 ;
   double epsilon1,epsilon2,epsilon4 ;
   /* == Variables ==
    *  kT : k * Te
    *  n_refrac : refractive index
    *  nu_c : cyclotron frequency
    *  nu_p : plasma frequency
    *  nuratio : nu/nu_c
    *  Tdl : dimensionless plasma temperature
    */

   kT = C_k*Te ;
   Tdl = kT/C_me_c2 ;
   nrhoe = rho/(C_me + C_mp) ;
   nu_c = C_e*B/(C_2pi*C_me*C_c) ;
   nu_p = C_e*sqrt(nrhoe/(C_me*C_pi)) ;
   nuratio = nu/nu_c ;
   sintheta = sin(theta) ;

   /* set gamma to kT: */
   gam = kT ;
   beta2 = 1.-gam*gam ;

   /* Eq.(0): */
   n_ref = sqrt(1.-nu_p*nu_p/(nu*nu)) ;
   /* Eq.(24): */
   xi = sqrt(1.-n_ref*n_ref*beta2) ;
   /* Eq.(36): */
   r = nuratio*gam/(1.5*xi*xi*sintheta) ;

   /* coefficient in Eq.(37) */
   coeff = (sqrt(3.)*C_e*C_e*nu_c*xi*sintheta)/(8.*C_pi*C_c*gam) ;

   rInv = 1./r ;
   r2Inv = rInv*rInv ;
   expMinusR = exp(-r) ;
   sqrtR = sqrt(r) ;
   sqrtRInv = 1./sqrtR ;
   r2 = r*r ;
   r2_3 = pow(r2,0.333333333333333333) ;
   r4_3 = r2_3*r2_3 ;

   /* epsilon_1: */
   epsilon1 = coeff * 2.*sqrtR*expMinusR*(1.2533141373+0.95739274378*rInv-1.2270825432*r2Inv) ;
   /* epsilon_2: */
   epsilon2 = coeff * 2.*r2_3*(1.0747641208-1.2657191442/r4_3+0.8060730906*r2Inv) ;
   /* epsilon_4: */
   epsilon4 = coeff * 2.66666666666666666/(tan(theta)*xi)
              * ( expMinusR*sqrtRInv*(1.2533141373-0.50480708309*rInv+0.70220889503*r2Inv)
                + (1.6876255256/r2_3-1.6121461812/r4_3+0.6328595721*r2Inv) ) ;

   /* jnu_i: */
   *jnuI = 0.5*nrhoe * C_2pi*C_me_c2*epsilon1 ;
   *jnuQ = 0.5*nrhoe * C_2pi*C_me_c2*epsilon2 ;
   *jnuV = 0.5*nrhoe * C_2pi*C_me_c2*epsilon4 ;

   /* using Kirchoff's Law to find j_nu */
   Bnu = 2.*C_h*nu*nu*nu / ( C_c2 * (exp(C_h*nu/kT)-1.)) ;
   *alphaI = *jnuI / Bnu ;
   *alphaQ = *jnuQ / Bnu ;
   *alphaV = *jnuV / Bnu ;
}
/* ------ for T9EQ -------------------------------------------------------}}} */

/* {{{--- common for all models --------------------------------------------- */
/* Blackbody intensity */
double BnuFun(double nu, double Te)
{
   return(2.*C_h*nu*nu*nu / ( C_c2 * (exp(C_h*nu/C_k/Te)-1.)) ) ;
}

/* post-call after calling any other subroutine in emiss.c */
int post_emiss()
{
#if CUBIC_SPLINE_INTERP
   free_dvector(a_Fx_splint,1,n_Fx_Gx) ;
   free_dvector(a_Fx2_splint,1,n_Fx_Gx) ;
 #if INTERPOLATE_RATIO
   free_dvector(a_GxOVERFx_splint,1,n_Fx_Gx) ;
   free_dvector(a_GxOVERFx2_splint,1,n_Fx_Gx) ;
 #else /* if INTERPOLATE_RATIO */
   free_dvector(a_Gx_splint,1,n_Fx_Gx) ;
   free_dvector(a_Gx2_splint,1,n_Fx_Gx) ;
 #endif /* if INTERPOLATE_RATIO */
 #if LOG10_INTERPOLATION
   free_dvector(log10x_Fx_Gx_splint,1,n_Fx_Gx) ;
 #else
   free_dvector(x_Fx_Gx_splint,1,n_Fx_Gx) ;
 #endif
#endif /* if CUBIC_SPLINE_INTERP */
   return(0) ;
}
/* ------ common for all models ------------------------------------------}}} */


/*******************************************************************************/
/*******************************************************************************
  jnu_thermal() :
 -----------------
    -- calculates the emissivity assuming that the matter is in thermal 
        equilibrium, given some cross section "sigma", material number 
        density "ne", 
        frequency of light "nu", and material temperature "Temp"
*******************************************************************************/
double jnu_thermal( double sigma, double ne, double nu, double Temp ) 
{
  double x, A, t1, log_A_inv, A_inv, jnu, alpha;

  alpha = ne * sigma; 
  x =  C_h * nu / (C_k * Temp);   
  A = 2.*C_h*nu*nu*nu*alpha / C_c2;

  /* This is used to elimate overflow problems and to increase accuracy: */
  if( x  > 100. ) { 
    A_inv = 1. / A;
    log_A_inv = log( A_inv );
    if( log_A_inv < 0. ) { 
      jnu = 1. / ( exp( log_A_inv + x )  -  A_inv ) ;
    } 
    else { 
      jnu = A / ( exp( x )  -  1. );
    }
  }
  /* If it is small, then use a 9th-order Taylor expansion of the exponent : */
  else if ( x < 1.e-3 ) { 
    t1  = x*x;
    jnu = A*(239500800.0+(-119750400.0+(19958400.0+(-332640.0+(7920.0+(-198.0
      +5.0*t1)*t1)*t1)*t1)*x)*x)/(x*239500800.0);
  }
  /* Standard expression: */
  else { 
    jnu = A / ( exp( x )  -  1. );
  }

  return( jnu ) ; 

}

/*******************************************************************************/
/*******************************************************************************
  sigma_Klein_Nishina() :
 -----------------
    -- returns the Klein-Nishina cross section of photon scattering off of an 
       electron with energy transfer.  This represents the same situation 
       as "Thomson" scattering except that Thomson scattering entails no 
       energy transfer between the electron and photon. 

    -- "nu" is the frequency of the photon; 

    -- sigma_{KN} -> sigma_{thomson} as  eps->0  where eps = h nu / (m c^2)
        and 
       sigma_{KN} -> (3/8) sigma_{thomson}/eps * ( log(2*eps) + 0.5 ) 

        (Padmanabhan)

*******************************************************************************/
double sigma_Klein_Nishina( double nu ) 
{

  double eps, term, sigma, x, A, y; 
  double t0, t2, t3, t4, t6, t7;

  x = eps = C_h * nu / C_me_c2 ; 

  A = ( 3. * C_sigmaT ) / 8. ; 
  term = 2.*eps + 1.;

  if( eps < 1.e-2 ) { 
    sigma =  (8.0/3.0+(-16.0/3.0+(208.0/15.0+
        (-532.0/15.0+(9152.0/105.0+(-4352.0/21.0+(30272.0/63.0
    +(-49184.0/45.0+1212416.0/495.0*x)*x)*x)*x)*x)*x)*x)*x)*A ; 
  } 
  else if( eps > 1.e2 ) { 
    y = 1./x;
    t2 = log(2.0);
    t4 = log(y);
    t6 = 1290240.0*t2;
    t7 = 1290240.0*t4;
    sigma = -A*y*(-645120.0*t2+645120.0*t4-322560.0+(-2903040.0+t6-t7+(806400.0
        +t6-t7+(376320.0+(-36960.0+(-10752.0+(14784.0+(-11136.0+(7215.0+(-4370.0+2548.0
            *y)*y)*y)*y)*y)*y)*y)*y)*y)*y)/645120.0;
  }    
  else { 
    t0 = 2.*x + 1.;
    t3 = log(t0);
    t7 = 4.0*t3;
    sigma = A*(-2.0*t3+(-10.0*t3+4.0+(-15.0*t3+16.0+(-t7+18.0+(t7+2.0)*x)*x)*x)*x);
    sigma /= x*x*x*t0*t0;

//    sigma =  ( 1. - 2.*(1.+eps)/(eps*eps) ) * log(1. + 2.*eps) 
//      + 0.5 + 4./eps - 0.5/(term*term) ;
//    sigma *= A  ;

  }

  return( sigma ) ; 
}

/*******************************************************************************/
/*******************************************************************************
  jnu_C0() :
 -----------------
    -- emissivity of unpolarized, isotropic Compton scattering using 
       the Klein-Nishina cross-section that includes QED corrections.

*******************************************************************************/
double jnu_C0(double nu, double rho, double Te) 
{
  double nrhoe, sigma;

  nrhoe = rho/(C_me + C_mp) ;
  sigma = sigma_Klein_Nishina(nu) ;
  return( jnu_thermal( sigma, nrhoe, nu, Te ) ) ; 
}

/******************************************************************************/
/*                           Main subroutines                                 */
/******************************************************************************/

/*  specific emissivity of electron (in erg cm^-3 s^-1 ster^-1 Hz^-1)
 */
double jnu_approx(int emissioneq, double nu, double rho, 
           double Te, double B, double theta)
{
   double ans ;
   /* == Input variables =
    *  emissioneq : integer denoting different equation used
    *  nu : emitted frequency (in Hz) in the fluid frame
    *  rho : rest-mass density of the gas (in g cm^-3)
    *  Te : electron temperature (in Kevin)
    *  B : magnetic field strength (in Gauss)
    *  theta : the pitch angle (in radian)
    */
   /*  nuratio : nu / nu_c
    *  Tdl : dimensionless temperature (Te/(me c^2)
    *  lg_... : log10(...) */
   double nuratio,Tdl ;
   //double lg_nuratio,lg_Tdl ;

   switch( emissioneq ) { 
   case B0EQ   :    return(jnu_B0(nu,rho,Te)) ;
   case T0EQ   :    return(jnu_T0(nu,rho,Te,B,theta)) ;
   case T1EQ   :    return(jnu_T1(nu,rho,Te,B,theta)) ;
   case T2EQ   :    return(jnu_T2(nu,rho,Te,B,theta)) ;
   case T3EQ   :    return(jnu_T3(nu,rho,Te,B,theta)) ;
   case T4EQ   :    return(jnu_T4(nu,rho,Te,B,theta)) ;
   case T5EQ   :    return(jnu_T5(nu,rho,Te,B,theta)) ;
   case T6EQ   :    return(jnu_T6(nu,rho,Te,B,theta)) ;
   case C0EQ   :    return(jnu_C0(nu,rho,Te)); 
   case T0_B0EQ :
      ans = jnu_B0(nu,rho,Te) + jnu_T0(nu,rho,Te,B,theta) ;
      return(ans) ;
   case T1_B0EQ :
      ans = jnu_B0(nu,rho,Te) + jnu_T1(nu,rho,Te,B,theta) ;
      return(ans) ;
   case T2_B0EQ :
      ans = jnu_B0(nu,rho,Te) + jnu_T2(nu,rho,Te,B,theta) ;
      return(ans) ;
   case T3_B0EQ :
      ans = jnu_B0(nu,rho,Te) + jnu_T3(nu,rho,Te,B,theta) ;
      return(ans) ;
   case T4_B0EQ :
      ans = jnu_B0(nu,rho,Te) + jnu_T4(nu,rho,Te,B,theta) ;
      return(ans) ;
   case T5_B0EQ :
      ans = jnu_B0(nu,rho,Te) + jnu_T5(nu,rho,Te,B,theta) ;
      return(ans) ;
   case T6_B0EQ :
      ans = jnu_B0(nu,rho,Te) + jnu_T6(nu,rho,Te,B,theta) ;
      return(ans) ;
   default     : 
      
      fprintf(stderr,"ERROR: unknown emissioneq (=%3d)!\n",emissioneq) ;
      exit(1) ;
   }
   return(-1);
}

/* ========================================================================== */

/*  absorption coefficient of electron (in cm^-1), assuming thermal
 */
double alphanu_approx(int emissioneq, double nu, double rho, 
               double Te, double B, double theta)
{
   /* == Input variables =
    *  emissioneq : integer denoting different equation used
    *  nu : emitted frequency (in Hz) in the fluid frame
    *  rho : rest-mass density of the gas (in g cm^-3)
    *  Te : electron temperature (in Kevin)
    *  B : magnetic field strength (in Gauss)
    *  theta : the pitch angle (in radian)
    */

   /* assuming thermal equilibrium, so alpha_nu = j_nu / B_nu */
   switch( emissioneq ) { 
   default     : 
      return( ( jnu_approx(emissioneq,nu,rho,Te,B,theta) / BnuFun(nu,Te) ) ) ;
   }
}


/* vim:set ts=3 expandtab */
