/* Integrate specific emissivity j_\nu and absorptivity \alpha_\nu
 *
 * The output subroutines are
 *   jnu() : j_\nu
 *   alphanu() : \alpha_\nu
 *   jnu_average() : angle-averaged j_\nu
 *
 * The first argument is for choosing
 *   TOTAL : total emission
 *   MODE_O : ordinary mode (O)
 *   MODE_X : extraordinary mode (X)
 *   STOKES_I : Stokes parameter I
 *   STOKES_Q : Stokes parameter Q
 *   STOKES_V : Stokes parameter V
 */

/*----------------------------------------------------------------------------*/

/* According to Chanmugam et al. 89 Eq.(6) and Chanmugam & Dulk 81 Eq.(7),
 *  there is an extra factor of refractive index in \alpha_\nu compared to
 *  j_\nu. When n_+ and n_- are very different from each other, the ratio
 *  of the absorption coefficients (of o- and x-mode) would be different 
 *  from the ratio of emissivion coefficients (of o- and x-mode). Here,
 *  if REFRACTIVE is defined, the factor of refractive index will be 
 *  included in the calculation. */
/* This is not an important effect for our application, therefore I never
 *  really check this part of the code. And I am not sure whether it is 
 *  refractive index n or ray-refractive index n_r. In Chanmugam & Dulk 81 
 *  Eq.(7), the factor is shown as ray-refractive index. The relationship 
 *  between those two kinds of refractive indices is in Bekefi 1966 
 *  Eq.(1.121). */
#define REFRACTIVE
#undef REFRACTIVE

/*----------------------------------------------------------------------------*/
/* == header files and definitions == */

#include "decs.h"
#include "error.h"

static double jnu_integrand(double gamma, double tdl, double n, 
                            double nuratio, double cos_theta, 
                            double sin_theta);
static double jnuOX_integrand(int flag, double gamma, double tdl, 
                              double n, double nuratio, double cos_theta,
                              double sin_theta, double B, double nrhoe);
static double jnuOX_integrand_modified(int flag, double gamma, double tdl, 
                                       double n, double nuratio, double cos_theta,
                                       double sin_theta, double B, double nrhoe);
static double jnuStokes_integrand(int flag, double gamma, double tdl, 
                                  double n, double nuratio, 
                                  double cos_theta, double sin_theta);
static double jnuStokes_integrand_modified(int flag, double gamma, double tdl, 
                  double n, double nuratio, double cos_theta, double sin_theta);
static double alphanu_integrand(double gamma,
                                double tdl, double n, 
                                double nuratio, double cos_theta, 
                                double sin_theta, double B, 
                                double nrhoe);
static double alphanuOX_integrand(int flag, double gamma,
                                  double tdl, double n, 
                                  double nuratio, double cos_theta, 
                                  double sin_theta, double B, double nrhoe,
                                  int flag_Tx_interp, void *Tx_list);
static double alphanuStokes_integrand(int flag, double gamma,
                                      double tdl, double n, 
                                      double nuratio, double cos_theta, 
                                      double sin_theta, double B, 
                                      double nrhoe);

static double alphanufunc_g(double g, void *par);

static double ne(double gamma, double mu, double *ne_para);

#if 0
static void assign_par(TYPE_PAR_n *par_n, TYPE_PAR_jnu_I *par_Integration, 
                double (*integrand_g_fun)(double x,void*P),
                double nuratio, double nrhoe, double tdl, double B,
                double sin_theta, double cos_theta,
                size_t n_wksp,
                gsl_integration_workspace *wksp_g,
                gsl_integration_workspace *wksp_n,
                int flag_Tx_interp, void *Tx_list);
#endif

// /*============================================================================*/
// /* Dimensional numbers and coefficients in front of the result of integration.
//  * Emissivity = (this factor) * (result of n & g integrations) */
// static double extra_factor(double nrhoe, double tdlinv, 
//                                  double nu, double cos_theta)
// {
// #if (ELECTRON_DIST == THERMAL)
//         return(nrhoe*tdlinv*K2_tdlinv_inv(tdlinv)*C_2pi*C_e2*nu/(C_c*fabs(cos_theta)));
// #elif (ELECTRON_DIST == NONTHERMAL \
//     || ELECTRON_DIST == GAUSSIAN)
//         return(nrhoe*C_2pi*C_e2*nu/(C_c*fabs(cos_theta)));
// #else
//   #error other ELECTRON_DIST not yet implemented!
// #endif
// }

/*----------------------------------------------------------------------------*/
/* Dimensional numbers and coefficients in front of the result of integration.
 * Emissivity = (this factor) * (result of n & g integrations) */
//static double extra_factor_modified(double nrhoe, double tdlinv, double nu,
//                                 double cos_theta)
static double extra_factor_modified(double *para)
{
        double nrhoe = para[0];
        double nu = para[1];
        double cos_theta = para[2];
        return(nrhoe*C_e2*C_2pi*nu/(C_c*fabs(cos_theta)));
}

/*============================================================================*/
/* return argument of Bessel function */
static double argument_jn(double gam, double n, double nuratio,
                          double sin_theta, double cos_theta)
{
        double beta,n0,mu,sinxi;
        beta = sqrt(1. - 1./(gam*gam));
        n0 = nuratio * gam;
        mu = ( n0 - n ) / ( n0 * beta * cos_theta );
        sinxi = sqrt(1. - mu*mu);
        return( n0 * beta * sin_theta * sinxi );
}

/*============================================================================*/
/* Calculate boundaries of gamma integration */
/* When tmp2 is negative, there is no solution. No gamma is corresponded
 *  to the inputed n and parameters. The integral is there zero.
 * When tmp2 is 0, there is only 1 solution. The two limits of the gamma
 *  integral are the same, therefore the gamma integral is zero.
 * When tmp2 is positive, there are 2 solution. */
/* NO SOLUTION --> return 1
 * NORMAL EXIT --> return 0 */
static int return_g_boundaries(double n, double nuratio,
                               double sin_theta, double cos_theta, 
                               double *gBegin, double *gEnd)
{
        double tmp1,tmp2,tmp3,sin2,sin2inv,fabs_cos_theta;

        sin2 = sin_theta * sin_theta;
        sin2inv = 1./sin2;
        fabs_cos_theta = fabs( cos_theta );
        tmp1 = n / nuratio;
        tmp2 = tmp1 * tmp1 - sin2;
        if(tmp2<=0.) {
                *gBegin = *gEnd = 0.;
                return(1); /* no solution */
        }
        tmp3 = fabs_cos_theta * sqrt(tmp2);
        *gBegin = (tmp1 - tmp3) * sin2inv;
        if(*gBegin < 1.) *gBegin = 1.;
        *gEnd = (tmp1 + tmp3) * sin2inv;
        if(*gEnd < *gBegin) return(1); /* no solution */

#if (ELECTRON_DIST == THERMAL)
  #if (FLAG_GAMMA_BOUNDARY == 1)
        static double tmp4,gPeak,jnPeak,gB1,gE1,jnB1,jnE1;
        static int i;
        gPeak = tmp1 * sin2inv;
        jnPeak = my_Bessel_J(n, argument_jn(gPeak,n,nuratio,sin_theta,cos_theta));
        for(i=0;i<GAMMA_BOUNDARY_nr;i++) {
                tmp4 = tmp1*fabs_cos_theta*GAMMA_BOUNDARY_sqrt1mr2[i];
                gB1 = (tmp1 - tmp4) * sin2inv;
                if(gB1 < *gBegin) continue;
                jnB1 = my_Bessel_J(n, argument_jn(gB1,n,nuratio,sin_theta,cos_theta));
                if(jnB1<JN_RATIO*jnPeak) {
                        gE1 = (tmp1 + tmp4) * sin2inv;
                        jnE1 = my_Bessel_J(n, argument_jn(gE1,n,nuratio,sin_theta,cos_theta));
                        if(jnE1<JN_RATIO*jnPeak) {
                                if(gB1 > *gBegin && gB1 < *gEnd) *gBegin = gB1;
                                if(gE1 > *gBegin && gE1 < *gEnd) *gEnd = gE1;
                                break;
                        }
                }
        }
  #endif /* if (FLAG_GAMMA_BOUNDARY == 1) */
#elif (ELECTRON_DIST == GAUSSIAN)
        /* Gaussian function drops off quickly on both sides.
         * Play safe, limit to width = several DELTA on each side */
        double width=10.*DELTA;
        //double width=5.*DELTA;
        if(*gBegin < GAM0-width) *gBegin = GAM0-width;
        if(*gEnd   > GAM0+width) *gEnd   = GAM0+width;
        if(*gEnd < *gBegin) return(1); /* no solution */
#endif /* if (ELECTRON_DIST == ...) */

        return(0);
}

/*============================================================================*/
/* for total emissivity */
static double func_g(double g, void *par)
{
        TYPE_PAR_g *p = (TYPE_PAR_g *)par;

        return(jnuOX_integrand_modified(0,g,p->tdl,p->n,p->nuratio,p->cos_theta,
                              p->sin_theta,p->B,p->nrhoe));
}

/*----------------------------------------------------------------------------*/
/* for ordinary (O) mode */
static double funcO_g(double g, void *par)
{
        TYPE_PAR_g *p = (TYPE_PAR_g *)par;

        return(jnuOX_integrand_modified(1,g,p->tdl,p->n,p->nuratio,p->cos_theta,
                              p->sin_theta,p->B,p->nrhoe));
}

/*----------------------------------------------------------------------------*/
/* for extraordinary (X) mode */
static double funcX_g(double g, void *par)
{
        TYPE_PAR_g *p = (TYPE_PAR_g *)par;

        return(jnuOX_integrand_modified(-1,g,p->tdl,p->n,p->nuratio,p->cos_theta,
                              p->sin_theta,p->B,p->nrhoe));
}

/*----------------------------------------------------------------------------*/
/* for ordinary (O) mode */
static double funcO_g_modified(double g, void *par)
{
        TYPE_PAR_g *p = (TYPE_PAR_g *)par;

        return(jnuOX_integrand_modified(1,g,p->tdl,p->n,p->nuratio,p->cos_theta,
                              p->sin_theta,p->B,p->nrhoe));
}

/*----------------------------------------------------------------------------*/
/* for extraordinary (X) mode */
static double funcX_g_modified(double g, void *par)
{
        TYPE_PAR_g *p = (TYPE_PAR_g *)par;

        return(jnuOX_integrand_modified(-1,g,p->tdl,p->n,p->nuratio,p->cos_theta,
                              p->sin_theta,p->B,p->nrhoe));
}

/*----------------------------------------------------------------------------*/
/* for Stokes parameter I */
static double funcI_g(double g, void *par)
{
        TYPE_PAR_g *p = (TYPE_PAR_g *)par;

        return(jnuStokes_integrand_modified(1,g,p->tdl,p->n,p->nuratio,
                                   p->cos_theta,p->sin_theta));
}

/*----------------------------------------------------------------------------*/
/* for Stokes parameter Q */
static double funcQ_g(double g, void *par)
{
        TYPE_PAR_g *p = (TYPE_PAR_g *)par;

        return(jnuStokes_integrand_modified(2,g,p->tdl,p->n,p->nuratio,
                                   p->cos_theta,p->sin_theta));
}

/*----------------------------------------------------------------------------*/
/* for Stokes parameter V */
static double funcV_g(double g, void *par)
{
        TYPE_PAR_g *p = (TYPE_PAR_g *)par;

        return(jnuStokes_integrand_modified(4,g,p->tdl,p->n,p->nuratio,
                                   p->cos_theta,p->sin_theta));
}

/*----------------------------------------------------------------------------*/

static double func_n(double n, void *par)
{
        TYPE_PAR_n *p = (TYPE_PAR_n *)par;
        int status;
        double ans,gAbsErr;

        /* Emissivity j_\nu is zero when the electron is aligned or anti-aligned
         *  with the magentic field */
        if((p->par_g_pt)->sin_theta == 0.) return(0.);

#if INTEGRATE_WITH_LOGN == TRUE
        /* first argument is actually log(n) in this case */
        double logn = n;
        n = exp(logn);
#endif

        /* Pass n to GSL subroutine by using point *p */
        (p->par_g_pt)->n = n;

        /* Calculate boundaries of g */
        status = return_g_boundaries(n, (p->par_g_pt)->nuratio, 
                                           (p->par_g_pt)->sin_theta,
                                           (p->par_g_pt)->cos_theta,
                                           &(p->gBegin), &(p->gEnd));
        if(status) return(0.);

        gsl_set_error_handler_off();
#if INTEGRATOR == GSL_QAG
        status = gsl_integration_qag( p->integrand_g_pt, p->gBegin, p->gEnd, 
                                      p->gAbsEps, p->gRelEps, p->n_wksp_g, 
                                      GSL_QAG_RULE_g, p->wksp_g_pt,
                                      &ans, &gAbsErr );
#elif INTEGRATOR == GSL_QNG
        size_t neval;
        status = gsl_integration_qng( p->integrand_g_pt, p->gBegin, p->gEnd, 
                                      p->gAbsEps, p->gRelEps, &ans, &gAbsErr,
                                      &neval );
#else
  #error "INTEGRATOR is not set correctly"
#endif

        if(status) {
#if FLAG_ERR1 == ERR1_ON_SCREEN_EXIT
          fprintf(stderr,"GSL error in  func_n().  gsl_errno = %d\n",status);
          exit(-1);
#elif FLAG_ERR1 == ERR1_ON_SCREEN_CONT
                fprintf(stderr,"GSL error in  func_n().  gsl_errno = %d\n",status);
#elif FLAG_ERR1 == ERR1_TO_FILE_CONT
          fprintf(errorpt,"GSL error in  func_n().  gsl_errno = %d\n",status);
          fflush(errorpt);
#elif FLAG_ERR1 == ERR1_IGNORE
#else
  #error "FLAG_ERR1 value unknown"
#endif
        }
        if(p->gMaxAbsErr < gAbsErr) p->gMaxAbsErr = gAbsErr;

#if INTEGRATE_WITH_LOGN == TRUE
        return(ans*n); /* d(log(n)) = d(n)/n */
#elif INTEGRATE_WITH_LOGN == FALSE
        return(ans);
#else
  #error invalid value of INTEGRATE_WITH_LOGN
#endif
}

/*============================================================================*/
/* -- integrand (without constant factors) of the gamma integration, for a 
 * certain n.
 */
/*
 * Small tdl:
 *   (1/beta) ( coeffA^2 J^2 + coeffB^2 J'^2 ) gam
 *    sqrt(gam^2-1) Exp(-gam/tdl) Exp(1/tdl)    ------ (eq 1)
 * Large tdl:
 *   (1/beta) ( coeffA^2 J^2 + coeffB^2 J'^2 ) gam
 *    sqrt(gam^2-1) Exp(-gam/tdl)               ------ (eq 2)
 *
 * Depending on the value of K2_EQ, either (eq 1) or (eq 2) is returned.
 */
static double jnu_integrand(double gamma, double tdl, double n, 
                           double nuratio, double cos_theta, double sin_theta)
{
        double jn, arg_jn, d_jn, n_0, beta, mu, sinxi, tmp1, tmp2;

        if(gamma==1.) return(0.); /* if electron is at rest, there's no emission */

        n_0 = nuratio * gamma ;

        if(gamma < 1.) {
                fprintf(stderr,"ERROR in jnu_integrand(), gamma < 1\n");
                fprintf(stderr,"%.16e %.16e %.16e %.16e %.16e %.16e\n",
                               gamma,tdl,n,nuratio,cos_theta,sin_theta);
                exit(-1);
        }
        beta = sqrt( (gamma-1.)*(gamma+1.) ) / gamma;

        mu = ( n_0 - n ) / ( n_0 * beta * cos_theta );

        if(mu >= -1. && mu <= 1.) {
                sinxi = sqrt( 1. - mu*mu );
        } else {
#if FLAG_ERR2 == ERR2_ON_SCREEN_EXIT
                fprintf(stderr,"ERROR in jnu_integrand(), ");
                fprintf(stderr," \\mu (i.e., cos(\\xi)) out of bound\n");
                fprintf(stderr," mu,n_0,n,beta,cos_theta = ");
                fprintf(stderr,"%.16e %.16e %.16e %.16e %.16e\n",
                               mu,n_0,n,beta,cos_theta);
                exit(-1);
#elif FLAG_ERR2 == ERR2_ON_SCREEN_CONT
                fprintf(stderr,"ERROR in jnu_integrand(), ");
                fprintf(stderr," \\mu (i.e., cos(\\xi)) out of bound\n");
                fprintf(stderr," mu,n_0,n,beta,cos_theta = ");
                fprintf(stderr,"%.16e %.16e %.16e %.16e %.16e\n",
                               mu,n_0,n,beta,cos_theta);
                if(mu < -1.) {
                        mu = -1.;
                        sinxi = 0.;
                } else {
                        mu = 1.;
                        sinxi = 0.;
                }
#elif FLAG_ERR2 == ERR2_TO_FILE_CONT
                fprintf(errorfixedpt,"ERROR in jnu_integrand(), ");
                fprintf(errorfixedpt," \\mu (i.e., cos(\\xi)) out of bound\n");
                fprintf(errorfixedpt," mu,n_0,n,beta,cos_theta = ");
                fprintf(errorfixedpt,"%.16e %.16e %.16e %.16e %.16e\n",
                                     mu,n_0,n,beta,cos_theta);
                fflush(errorfixedpt);
                if(mu < -1.) {
                        mu = -1.;
                        sinxi = 0.;
                } else {
                        mu = 1.;
                        sinxi = 0.;
                }
#elif FLAG_ERR2 == ERR2_FIX_CONT
                if(mu < -1.) {
                        mu = -1.;
                        sinxi = 0.;
                } else {
                        mu = 1.;
                        sinxi = 0.;
                }
#elif FLAG_ERR2 == ERR2_IGNORE
#else
  #error "FLAG_ERR2 value unknown"
#endif
        }

        arg_jn = n_0 * beta * sin_theta * sinxi;

        /* Bessel function routine: */
        d_jn = my_Bessel_dJ(n, arg_jn, &jn);

        tmp1 = jn   * (cos_theta - beta * mu) / sin_theta;
        tmp2 = d_jn * beta * sinxi;

#if (ELECTRON_DIST == THERMAL)
 #if K2_EQ == K2_WZ_whole_eq
        if( tdl <= 0.65 ) {
          /* K2(1/tdl) ~ A exp(-1/tdl) for small tdl
           * For thermal distribution, ne(gamma) contains the factor 1/K2(1/tdl),
           * here we put the factor 1/exp(-1/tdl) in the integrand */
                return( exp_factor( ( tmp1*tmp1 + tmp2*tmp2 ) * gamma*gamma ,
                        ((1.-gamma)/tdl) ) ) ;
        } else {
                return( exp_factor( ( tmp1*tmp1 + tmp2*tmp2 ) * gamma*gamma ,
                        (-gamma/tdl) ) ) ;
        }
 #elif K2_EQ == K2_WZ_high_T
        return( exp_factor( ( tmp1*tmp1 + tmp2*tmp2 ) * gamma*gamma ,
                      (-gamma/tdl) ) ) ;
 #elif K2_EQ == K2_GSL
        /* K2(1/tdl) ~ A exp(-1/tdl) for small tdl
         * For thermal distribution, ne(gamma) contains the factor 1/K2(1/tdl),
         * here we put the factor 1/exp(-1/tdl) in the integrand */
        return( exp_factor( ( tmp1*tmp1 + tmp2*tmp2 ) * gamma*gamma ,
                      ((1.-gamma)/tdl) ) ) ;
 #elif K2_EQ == K2_GSL_interpolate
  #error "ERROR: K2_EQ == K2_GSL_interpolate not yet implemented"
 #else
  #error "K2_EQ not defined correctly"
 #endif
#elif (ELECTRON_DIST == NONTHERMAL)
        DISTRIBUTION_PARAMETER;
        return(ne(gamma,mu,&ne_para[0])*(tmp1*tmp1+tmp2*tmp2)/beta);
#elif (ELECTRON_DIST == GAUSSIAN)
        DISTRIBUTION_PARAMETER;
        return(ne(gamma,mu,&ne_para[0])*(tmp1*tmp1+tmp2*tmp2)/beta);
#else
 #error other ELECTRON_DIST not yet implemented
#endif
}

/*============================================================================*/
/* -- emissivity integrand (without constant factors) of the gamma integration, 
 * for a certain n. Only for one of Stokes I, Q, V.
 * Stokes I  -->  flagMODE = STOKES_I
 * Stokes Q  -->  flagMODE = STOKES_Q
 * Stokes V  -->  flagMODE = STOKES_V
 */
/* Vath & Chanmugam 1995 ApJS, 98, 295 */
/*
 * Small tdl:
 *   (1/beta) ( coeffA^2 J^2 + coeffB^2 J'^2 ) gam
 *    sqrt(gam^2-1) Exp(-gam/tdl) Exp(1/tdl)    ------ (eq 1)
 * Large tdl:
 *   (1/beta) ( coeffA^2 J^2 + coeffB^2 J'^2 ) gam
 *    sqrt(gam^2-1) Exp(-gam/tdl)               ------ (eq 2)
 *
 * Depending on the value of K2_EQ, either (eq 1) or (eq 2) is returned.
 *
 * Note that coeffA and coeffB are not the same as those in the total 
 *  emissivity. However, they DO NOT include the index of refraction and 
 *  polarization coefficients (Ramaty 1969, Eq.(1),(2)).
 */
static double jnuStokes_integrand(int flagMODE, double gamma, double tdl, 
                                  double n, double nuratio, double cos_theta, 
                                  double sin_theta)
{
        double jn, arg_jn, d_jn, n_0, beta, mu, sinxi, tmp, tmp1, tmp2;

        if(gamma==1.) return(0.); /* if electron is at rest, there's no emission */

        n_0 = nuratio * gamma ;

        if(gamma < 1.) {
                fprintf(stderr,"ERROR in jnuStokes_integrand(), gamma < 1\n");
                fprintf(stderr,"%.16e %.16e %.16e %.16e %.16e %.16e\n",
                               gamma,tdl,n,nuratio,cos_theta,sin_theta);
                exit(-1);
        }
        beta = sqrt( (gamma-1.)*(gamma+1.) ) / gamma;

        mu = ( n_0 - n ) / ( n_0 * beta * cos_theta );

        if(mu >= -1. && mu <= 1.) {
                sinxi = sqrt( 1. - mu*mu );
        } else {
#if FLAG_ERR2 == ERR2_ON_SCREEN_EXIT
                fprintf(stderr,"ERROR in jnuStokes_integrand(), ");
                fprintf(stderr,"\\mu (i.e., cos(\\xi)) out of bound\n");
                fprintf(stderr," mu,n_0,n,beta,cos_theta = ");
                fprintf(stderr,"%.16e %.16e %.16e %.16e %.16e\n",mu,n_0,n,beta,cos_theta);
                exit(-1);
#elif FLAG_ERR2 == ERR2_ON_SCREEN_CONT
                fprintf(stderr,"ERROR in jnuStokes_integrand(), ");
                fprintf(stderr,"\\mu (i.e., cos(\\xi)) out of bound\n");
                fprintf(stderr," mu,n_0,n,beta,cos_theta = ");
                fprintf(stderr,"%.16e %.16e %.16e %.16e %.16e\n",mu,n_0,n,beta,cos_theta);
                if(mu < -1.) {
                        mu = -1.;
                        sinxi = 0.;
                } else {
                        mu = 1.;
                        sinxi = 0.;
                }
#elif FLAG_ERR2 == ERR2_TO_FILE_CONT
                fprintf(errorfixedpt,"ERROR in jnuStokes_integrand(), ");
                fprintf(errorfixedpt,"\\mu (i.e., cos(\\xi)) out of bound\n");
                fprintf(errorfixedpt," mu,n_0,n,beta,cos_theta = ");
                fprintf(errorfixedpt,"%.16e %.16e %.16e %.16e %.16e\n",
                                     mu,n_0,n,beta,cos_theta);
                fflush(errorfixedpt);
                if(mu < -1.) {
                        mu = -1.;
                        sinxi = 0.;
                } else {
                        mu = 1.;
                        sinxi = 0.;
                }
#elif FLAG_ERR2 == ERR2_FIX_CONT
                if(mu < -1.) {
                        mu = -1.;
                        sinxi = 0.;
                } else {
                        mu = 1.;
                        sinxi = 0.;
                }
#elif FLAG_ERR2 == ERR2_IGNORE
#else
  #error "FLAG_ERR2 value unknown"
#endif
        }

        arg_jn = n_0 * beta * sin_theta * sinxi;

        /* Bessel function routine: */
        d_jn = my_Bessel_dJ(n, arg_jn, &jn);

        tmp1 = jn   * (cos_theta - beta * mu) / sin_theta;
        tmp2 = d_jn * beta * sinxi;

        if(flagMODE == STOKES_I) {
                tmp = tmp1*tmp1 + tmp2*tmp2;
        } else if(flagMODE == STOKES_Q) {
                tmp = tmp1*tmp1 - tmp2*tmp2;
        } else if(flagMODE == STOKES_V) {
                tmp = -2*tmp1*tmp2;
        } else {
                fprintf(stderr,"ERROR in jnuStokes_integrand(), ");
                fprintf(stderr,"flagMODE=%d is not a valid value.\n",flagMODE);
                exit(-1);
        }

#if K2_EQ == K2_WZ_whole_eq
        if( tdl <= 0.65 ) {
                /* K2(1/tdl) ~ A exp(-1/tdl) for small tdl
                 * For thermal distribution, ne(gamma) contains the factor 1/K2(1/tdl),
                 * here we put the factor 1/exp(-1/tdl) in the integrand */
                return( exp_factor( tmp * gamma*gamma , ((1.-gamma)/tdl) ) ) ;
        } else {
                return( exp_factor( tmp * gamma*gamma , (-gamma/tdl) ) ) ;
        }
#elif K2_EQ == K2_WZ_high_T
        return( exp_factor( tmp * gamma*gamma , (-gamma/tdl) ) ) ;
#elif K2_EQ == K2_GSL
        /* K2(1/tdl) ~ A exp(-1/tdl) for small tdl
         * For thermal distribution, ne(gamma) contains the factor 1/K2(1/tdl),
         * here we put the factor 1/exp(-1/tdl) in the integrand */
        return( exp_factor( tmp * gamma*gamma , ((1.-gamma)/tdl) ) ) ;
#elif K2_EQ == K2_GSL_interpolate
  #error "ERROR: K2_EQ == K2_GSL_interpolate not yet implemented"
#else
  #error "K2_EQ not defined correctly"
#endif
}
/*----------------------------------------------------------------------------*/
/* -- emissivity integrand (without constant factors) of the gamma integration, 
 * for a certain n. Only for one of Stokes I, Q, V.
 * Stokes I  -->  flagMODE = STOKES_I
 * Stokes Q  -->  flagMODE = STOKES_Q
 * Stokes V  -->  flagMODE = STOKES_V
 */
/* Vath & Chanmugam 1995 ApJS, 98, 295 */
/*
 * Small tdl:
 *   (1/beta) ( coeffA^2 J^2 + coeffB^2 J'^2 ) gam
 *    sqrt(gam^2-1) Exp(-gam/tdl) Exp(1/tdl)    ------ (eq 1)
 * Large tdl:
 *   (1/beta) ( coeffA^2 J^2 + coeffB^2 J'^2 ) gam
 *    sqrt(gam^2-1) Exp(-gam/tdl)               ------ (eq 2)
 *
 * Depending on the value of K2_EQ, either (eq 1) or (eq 2) is returned.
 *
 * Note that coeffA and coeffB are not the same as those in the total 
 *  emissivity. However, they DO NOT include the index of refraction and 
 *  polarization coefficients (Ramaty 1969, Eq.(1),(2)).
 */
static double jnuStokes_integrand_modified(int flagMODE, double gamma, double tdl, 
                                           double n, double nuratio, double cos_theta, double sin_theta)
{
        double jn, arg_jn, d_jn, n_0, beta, mu, sinxi, tmp, tmp1, tmp2;

        if(gamma==1.) return(0.); /* if electron is at rest, there's no emission */

        n_0 = nuratio * gamma ;

        if(gamma < 1.) {
                fprintf(stderr,"ERROR in jnuStokes_integrand_modified(), gamma < 1\n");
                fprintf(stderr,"%.16e %.16e %.16e %.16e %.16e %.16e\n",
                               gamma,tdl,n,nuratio,cos_theta,sin_theta);
                exit(-1);
        }
        beta = sqrt( (gamma-1.)*(gamma+1.) ) / gamma;

        mu = ( n_0 - n ) / ( n_0 * beta * cos_theta );

        if(mu >= -1. && mu <= 1.) {
                sinxi = sqrt( 1. - mu*mu );
        } else {
#if FLAG_ERR2 == ERR2_ON_SCREEN_EXIT
                fprintf(stderr,"ERROR in jnuStokes_integrand_modified(), ");
                fprintf(stderr,"\\mu (i.e., cos(\\xi)) out of bound\n");
                fprintf(stderr," mu,n_0,n,beta,cos_theta = ");
                fprintf(stderr,"%.16e %.16e %.16e %.16e %.16e\n",mu,n_0,n,beta,cos_theta);
                exit(-1);
#elif FLAG_ERR2 == ERR2_ON_SCREEN_CONT
                fprintf(stderr,"ERROR in jnuStokes_integrand_modified(), ");
                fprintf(stderr,"\\mu (i.e., cos(\\xi)) out of bound\n");
                fprintf(stderr," mu,n_0,n,beta,cos_theta = ");
                fprintf(stderr,"%.16e %.16e %.16e %.16e %.16e\n",mu,n_0,n,beta,cos_theta);
                if(mu < -1.) {
                        mu = -1.;
                        sinxi = 0.;
                } else {
                        mu = 1.;
                        sinxi = 0.;
                }
#elif FLAG_ERR2 == ERR2_TO_FILE_CONT
                fprintf(errorfixedpt,"ERROR in jnuStokes_integrand_modified(), ");
                fprintf(errorfixedpt,"\\mu (i.e., cos(\\xi)) out of bound\n");
                fprintf(errorfixedpt," mu,n_0,n,beta,cos_theta = ");
                fprintf(errorfixedpt,"%.16e %.16e %.16e %.16e %.16e\n",
                                     mu,n_0,n,beta,cos_theta);
                fflush(errorfixedpt);
                if(mu < -1.) {
                        mu = -1.;
                        sinxi = 0.;
                } else {
                        mu = 1.;
                        sinxi = 0.;
                }
#elif FLAG_ERR2 == ERR2_FIX_CONT
                if(mu < -1.) {
                        mu = -1.;
                        sinxi = 0.;
                } else {
                        mu = 1.;
                        sinxi = 0.;
                }
#elif FLAG_ERR2 == ERR2_IGNORE
#else
  #error "FLAG_ERR2 value unknown"
#endif
        }

        arg_jn = n_0 * beta * sin_theta * sinxi;

        /* Bessel function routine: */
        d_jn = my_Bessel_dJ(n, arg_jn, &jn);

        tmp1 = jn   * (cos_theta - beta * mu) / sin_theta;
        tmp2 = d_jn * beta * sinxi;

        if(flagMODE == STOKES_I) {
                tmp = tmp1*tmp1 + tmp2*tmp2;
        } else if(flagMODE == STOKES_Q) {
                tmp = tmp1*tmp1 - tmp2*tmp2;
        } else if(flagMODE == STOKES_V) {
                tmp = -2*tmp1*tmp2;
        } else {
                fprintf(stderr,"ERROR in jnuStokes_integrand_modified(), ");
                fprintf(stderr,"flagMODE=%d is not a valid value.\n",flagMODE);
                exit(-1);
        }

        DISTRIBUTION_PARAMETER;
        return(tmp*ne(gamma,mu,&ne_para[0])/beta);
}

/*============================================================================*/
static void return_n2_a_theta(double *n2, double *a_theta,
                              int flag, double nuratio, double nrhoe, 
                              double B, double cos_theta, double sin_theta)
/* n2 is the square of index of refraction
 * Ramaty 1969 ApJ, 158, 753, Eq.(1) */
/* a_theta is the polarization coefficients
 * Ramaty 1969 ApJ, 158, 753, Eq.(2) */
/* Ordinary (O) mode      -->  flag = +1
 * Extrardinary (X) mode  -->  flag = -1 */
{
        double sin2,nuratio2,nu_pB2,nu2diff,tmp;

        sin2 = sin_theta*sin_theta;
        /* (\nu / \nu_B)^2: */
        nuratio2 = nuratio*nuratio;
        /* (\nu_p / \nu_B)^2: */
        nu_pB2 = C_4pi*nrhoe*C_me_c2/(B*B);
        /* (nu_p^2 - \nu^2) / \nu_B^2: */
        nu2diff = nu_pB2 - nuratio2;

        tmp = flag * nuratio * sqrt( nuratio2*sin2*sin2
                                     + 4.*nu2diff*nu2diff*cos_theta*cos_theta )
              - nuratio2*sin2;

        /* Eq.(1): */
        *n2 = 1. 
              + 2.*nu_pB2*nu2diff / ( tmp - 2.*nuratio2*nu2diff);
        /* Eq.(2): */
        *a_theta = -2.*nuratio*nu2diff*cos_theta/tmp;
}

/*----------------------------------------------------------------------------*/

/* return Axial ratio T of the polarization ellipse,
 * and also longitudinal component L of the polarization vector */
/* Ordinary (O) mode      -->  flag = +1
 * Extrardinary (X) mode  -->  flag = -1 */
/* T:
 *   Melrose & McPhedran 1991 Eq.(13.38)
 *   same as a_theta_- in Ramaty 1969 Eq.(2)
 *   P.S. seems that Broderick seems to use T_o in Broderick & Blandford 2004 for sin(\chi) */
/* L:
 *   Melrose & McPhedran 1991 Eq.(13.37)
 *   see as a_k_- in Ramaty 1969 Eq.(3) */
void return_T_L(double *T, double *L, int flag, double nuratio,
                double nrhoe, double B, double costheta, double sintheta)
{
        double nup2,nuc,nu,nu2,X,Y,Y2,Y2sintheta2,Delta;

        nup2 = nrhoe*C_e2/(C_pi*C_me);
        nuc = C_e*B/(C_2pi*C_me*C_c);
        nu = nuratio*nuc;
        nu2 = nu*nu;
        /* X is (nu_p/nu)^2, Y is nu_c/nu */
        X = nup2/nu2;
        Y = 1./nuratio;
        Y2 = Y*Y;
        Y2sintheta2 = Y2*sintheta*sintheta;
        /* Melrose & McPhedran 1991 Eq.(13.39) */
        Delta = sqrt(0.25*Y2sintheta2*Y2sintheta2+(1-X)*(1-X)*Y2*costheta*costheta);
        /* Melrose & McPhedran 1991 Eq.(13.38) */
        *T = (Y*(1-X)*costheta) / (0.5*Y2*sintheta*sintheta - flag*Delta);

        /* Melrose & McPhedran 1991 Eq.(13.37) */
        *L = X*Y*sintheta*(*T)/((1.-X)*((*T)-Y*costheta));
}

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

/* return a table of accurate T_x as a function of nu,
 * calculated by using direct intergration of alphanu_Q and alphanu_V */
/* pre_... is to be called before ..., and pro_... is to be called after ... */
void pre_Tx_interp(void *Tx_list, int length, double nu1, double nu2, 
                   double rho, double Te, double B, double theta)
{
        TYPE_Tx *p = (TYPE_Tx *)Tx_list;
        double alphanuQ, alphanuV, r_nu;
        int i;
        void *list = NULL;

        p->length = length;

        p->rho   = rho;
        p->Te    = Te;
        p->B     = B;
        p->theta = theta;
        p->nuc   = C_e*B/(C_2pi*C_me*C_c);

        /* allocate memory to vector nu and Tx */
        if( (p->nu = (double *) malloc(length * sizeof(double))) == NULL) {
                fprintf(stderr,"Failed to allocate space for x in pre_interpolation()\n");
                exit(1);
        }
        if( (p->Tx = (double *) malloc(length * sizeof(double))) == NULL) {
                fprintf(stderr,"Failed to allocate space for x in pre_interpolation()\n");
                exit(1);
        }

        /* allocate memory for gamma-n integration */
        gsl_integration_workspace *wksp_g = gsl_integration_workspace_alloc(N_WKSP);
        gsl_integration_workspace *wksp_n = gsl_integration_workspace_alloc(N_WKSP);

        r_nu = nu2 / nu1;
        for(i=0;i<length;i++) {
                p->nu[i] = nu1 * pow(r_nu , i/((double)(length - 1)));
                alphanuQ = alphanu(STOKES_Q, p->nu[i], rho, Te, B, theta,
                                     N_WKSP, wksp_g, wksp_n, 0, list);
                alphanuV = alphanu(STOKES_V, p->nu[i], rho, Te, B, theta,
                                     N_WKSP, wksp_g, wksp_n, 0, list);

                /* Using these notations:
                 * alphanu_Q = 0.5 cos(2 chi) (alphanu_O - alphanu_X)
                 * alphanu_V = 0.5 sin(2 chi) (alphanu_O - alphanu_X) */
                p->Tx[i] = tan(0.5 * atan2(alphanuV,alphanuQ));
        }

        /* allocate memory to arrays used during interpolation */
        p->acc = gsl_interp_accel_alloc();
        p->spline = gsl_spline_alloc(gsl_interp_cspline,length);

        /* initialization */
        gsl_spline_init(p->spline,p->nu,p->Tx,length);

        /* free memory used for gamma-n integration */
        gsl_integration_workspace_free(wksp_g);
        gsl_integration_workspace_free(wksp_n);
}

/*----------------------------------------------------------------------------*/

/* return accurate T_x as a function of nu
 * by using interpolation of precalculated T_x,
 * assumptions:
 * 1) rho, Te, B and theta are assumed to be the same as 
 *    those used when pre_Tx_interp() was called.
 * 2) nu1 <= nu <= nu2 */
double Tx_interp(void *Tx_list, double nu)
{
        TYPE_Tx *p = (TYPE_Tx *)Tx_list;

        /* sanity check */
        if(nu < p->nu[0] || nu > p->nu[p->length - 1]) {
                fprintf(stderr,"ERROR in Tx_interp(), nu is out of range\n");
                fprintf(stderr,"nu = %.15g   is outside (%.15g,%.15g)\n",nu,p->nu[0],p->nu[p->length - 1]);
                exit(-1);
        }

        return( gsl_spline_eval(p->spline, nu, p->acc) );

}

/*----------------------------------------------------------------------------*/

/* Clean up after interpolation */
/* pre_... is to be called before ..., and pro_... is to be called after ... */
void pro_Tx_interp(void *Tx_list)
{
        TYPE_Tx *p = (TYPE_Tx *)Tx_list;

        /* free memory of arrays nu and Tx */
        free(p->nu);
        free(p->Tx);

        /* free memory of arrays used during interpolation: */
}

/*============================================================================*/
/* -- integrand (without constant factors) of the gamma integration, for a 
 * certain n. Only one of the two modes.
 * Ordinary (O) mode      -->  flag = +1
 * Extrardinary (X) mode  -->  flag = -1
 */
/* Ramaty 1969 ApJ, 158, 753, Eq.(14), with Eq.(7), Eq.(1) and Eq.(2).
 * Also see Chanmugam & Dulk 1981 ApJ, 244, 569, Eq.(5), which does the 
 *  integration in p and \phi instead of \gamma and \phi.
 * Also see Stupp 2000 MNRAS, 311, 251, Eq.(12), which does not contain the
 *  factor 1+d(log(n_ref))/d(log(\nu)), and contains the longitudinal 
 *  contribution, and keeps a pitch angle dependence.
 */
/*
 * Small tdl:
 *   (1/beta) ( coeffA^2 J^2 + coeffB^2 J'^2 ) gam
 *    sqrt(gam^2-1) Exp(-gam/tdl) Exp(1/tdl)    ------ (eq 1)
 * Large tdl:
 *   (1/beta) ( coeffA^2 J^2 + coeffB^2 J'^2 ) gam
 *    sqrt(gam^2-1) Exp(-gam/tdl)               ------ (eq 2)
 *
 * Depending on the value of K2_EQ, either (eq 1) or (eq 2) is returned.
 *
 * Note that coeffA and coeffB are not the same as those in the total 
 *  emissivity. Here they include the index of refraction and polarization
 *  coefficients (Ramaty 1969, Eq.(1),(2)).
 */
static double jnuOX_integrand(int flag, double gamma, double tdl, 
                              double n, double nuratio, double cos_theta, 
                              double sin_theta, double B, double nrhoe)
{
        double jn, d_jn, arg_jn, n_0, beta, mu, sinxi, tmp, tmp1, tmp2;

        if(gamma==1.) return(0.); /* if electron is at rest, there's no emission */

        /* Ramaty 1969 Eq.(1),(2) */
#define ORIGINAL
#undef ORIGINAL
#ifdef ORIGINAL
        double n2;
        return_n2_a_theta(&n2, &a_theta, flag, nuratio, nrhoe, B, cos_theta, sin_theta);
        n_refrac = sqrt(n2);
#else /* ifdef ORIGINAL */
        double T, L;
        return_T_L(&T, &L, -1, nuratio, nrhoe, B, cos_theta, sin_theta);
#endif /* ifdef ORIGINAL */

        n_0 = nuratio * gamma ;

        if(gamma < 1.) {
                fprintf(stderr,"ERROR in jnuOX_integrand(), gamma < 1\n");
                fprintf(stderr,"%.16e %.16e %.16e %.16e %.16e %.16e\n",
                               gamma,tdl,n,nuratio,cos_theta,sin_theta);
                exit(-1);
        }
        beta = sqrt( (gamma-1.)*(gamma+1.) ) / gamma;

#ifdef ORIGINAL
        mu = ( n_0 - n ) / ( n_refrac * n_0 * beta * cos_theta );
#else /* ifdef ORIGINAL */
        mu = ( n_0 - n ) / ( n_0 * beta * cos_theta );
#endif /* ifdef ORIGINAL */

        if(mu >= -1. && mu <= 1.) {
                sinxi = sqrt( 1. - mu*mu );
        } else {
#if FLAG_ERR2 == ERR2_ON_SCREEN_EXIT
                fprintf(stderr,"ERROR in jnuOX_integrand(), ");
                fprintf(stderr,"\\mu (i.e., cos(\\xi)) out of bound\n");
                fprintf(stderr," mu,n_0,n,beta,cos_theta = ");
                fprintf(stderr,"%.16e %.16e %.16e %.16e %.16e\n",mu,n_0,n,beta,cos_theta);
                exit(-1);
#elif FLAG_ERR2 == ERR2_ON_SCREEN_CONT
                fprintf(stderr,"ERROR in jnuOX_integrand(), ");
                fprintf(stderr,"\\mu (i.e., cos(\\xi)) out of bound\n");
                fprintf(stderr," mu,n_0,n,beta,cos_theta = ");
                fprintf(stderr,"%.16e %.16e %.16e %.16e %.16e\n",mu,n_0,n,beta,cos_theta);
                if(mu < -1.) {
                        mu = -1.;
                        sinxi = 0.;
                } else {
                        mu = 1.;
                        sinxi = 0.;
                }
#elif FLAG_ERR2 == ERR2_TO_FILE_CONT
                fprintf(errorfixedpt,"ERROR in jnuOX_integrand(), ");
                fprintf(errorfixedpt,"\\mu (i.e., cos(\\xi)) out of bound\n");
                fprintf(errorfixedpt," mu,n_0,n,beta,cos_theta = ");
                fprintf(errorfixedpt,"%.16e %.16e %.16e %.16e %.16e\n",
                                     mu,n_0,n,beta,cos_theta);
                fflush(errorfixedpt);
                if(mu < -1.) {
                        mu = -1.;
                        sinxi = 0.;
                } else {
                        mu = 1.;
                        sinxi = 0.;
                }
#elif FLAG_ERR2 == ERR2_FIX_CONT
                if(mu < -1.) {
                        mu = -1.;
                        sinxi = 0.;
                } else {
                        mu = 1.;
                        sinxi = 0.;
                }
#elif FLAG_ERR2 == ERR2_IGNORE
#else
  #error "FLAG_ERR2 value unknown"
#endif
        }

#ifdef ORIGINAL
        /* Ramaty 1969 Eq.(8) (seems to be wrong, see Chanmugam & Dulk 81 Eq.(3) */
        //arg_jn = n_0 * n_refrac * beta * sin_theta * sinxi
        //         / (1. - n_refrac * beta * cos_theta * mu);
        arg_jn = n_0 * n_refrac * beta * sin_theta * sinxi;

        /* Bessel function routine: */
        d_jn = my_Bessel_dJ(n, arg_jn, &jn);

        tmp1 = jn   * a_theta * (cos_theta / n_refrac - beta * mu) / sin_theta;
        tmp2 = d_jn * beta * sinxi;

        /* Ramaty 1969 Eq.(7), second line (without square) */
        tmp = tmp1 - tmp2;
        /* Ramaty 1969 Eq.(7), first line (after summation sign), and second line */
        tmp = tmp * tmp * n_refrac / (1. + a_theta*a_theta);
#else /* ifdef ORIGINAL */
        /* Ramaty 1969 Eq.(8) (seems to be wrong, see Chanmugam & Dulk 81 Eq.(3) */
        arg_jn = n_0 * beta * sin_theta * sinxi;

        /* Bessel function routine: */
        d_jn = my_Bessel_dJ(n, arg_jn, &jn);

        if(flag == 1) {
                tmp1 =  jn   * (cos_theta - beta * mu) / sin_theta;
                tmp2 = -d_jn * T * beta * sinxi;
        } else if(flag == -1) {
                tmp1 = jn   * T * (cos_theta - beta * mu) / sin_theta;
                tmp2 = d_jn * beta * sinxi;
        } else { /* shouldn't have arrive here */
                fprintf(stderr,"ERROR in jnuOX_integrand(), flag=%d is not allowed\n",flag);
                exit(1);
        }

        /* Ramaty 1969 Eq.(7), second line (without square) */
        tmp = tmp1 + tmp2;
        /* Ramaty 1969 Eq.(7), first line (after summation sign), and second line */
        tmp = tmp * tmp / (1. + T*T);
#endif /* ifdef ORIGINAL */

#if K2_EQ == K2_WZ_whole_eq
        if( tdl <= 0.65 ) {
                /* K2(1/tdl) ~ A exp(-1/tdl) for small tdl
                 * For thermal distribution, ne(gamma) contains the factor 1/K2(1/tdl),
                 * here we put the factor 1/exp(-1/tdl) in the integrand */
  #ifdef REFRACTIVE
                return( exp_factor( ( tmp / n_refrac ) * gamma*gamma , ((1.-gamma)/tdl) ) ) ;
  #else /* ifdef REFRACTIVE */
                return( exp_factor( tmp * gamma*gamma , ((1.-gamma)/tdl) ) ) ;
  #endif /* ifdef REFRACTIVE */
        }
        else {
  #ifdef REFRACTIVE
                return( exp_factor( ( tmp / n_refrac ) * gamma*gamma , (-gamma/tdl) ) ) ;
  #else /* ifdef REFRACTIVE */
                return( exp_factor( tmp * gamma*gamma , (-gamma/tdl) ) ) ;
  #endif /* ifdef REFRACTIVE */
        }
#elif K2_EQ == K2_WZ_high_T
  #ifdef REFRACTIVE
        return( exp_factor( ( tmp / n_refrac ) * gamma*gamma , (-gamma/tdl) ) ) ;
  #else /* ifdef REFRACTIVE */
        return( exp_factor( tmp * gamma*gamma , (-gamma/tdl) ) ) ;
  #endif /* ifdef REFRACTIVE */
#elif K2_EQ == K2_GSL
        /* K2(1/tdl) ~ A exp(-1/tdl) for small tdl
         * For thermal distribution, ne(gamma) contains the factor 1/K2(1/tdl),
         * here we put the factor 1/exp(-1/tdl) in the integrand */
  #ifdef REFRACTIVE
        return( exp_factor( ( tmp / n_refrac ) * gamma*gamma , ((1.-gamma)/tdl) ) ) ;
  #else /* ifdef REFRACTIVE */
        return( exp_factor( tmp * gamma*gamma , ((1.-gamma)/tdl) ) ) ;
  #endif /* ifdef REFRACTIVE */
#elif K2_EQ == K2_GSL_interpolate
  #error "ERROR: K2_EQ == K2_GSL_interpolate not yet implemented"
#else
  #error "K2_EQ not defined correctly"
#endif
}

/*============================================================================*/
/* -- integrand (without constant factors) of the gamma integration, for a 
 * certain n. Only one of the two modes.
 * Ordinary (O) mode      -->  flag = +1
 * Extrardinary (X) mode  -->  flag = -1
 * O mode + X mode        -->  flag = 0
 */
/* Ramaty 1969 ApJ, 158, 753, Eq.(14), with Eq.(7), Eq.(1) and Eq.(2).
 * Also see Chanmugam & Dulk 1981 ApJ, 244, 569, Eq.(5), which does the 
 *  integration in p and \phi instead of \gamma and \phi.
 * Also see Stupp 2000 MNRAS, 311, 251, Eq.(12), which does not contain the
 *  factor 1+d(log(n_ref))/d(log(\nu)), and contains the longitudinal 
 *  contribution, and keeps a pitch angle dependence.
 */
/*
 * Small tdl:
 *   (1/beta) ( coeffA^2 J^2 + coeffB^2 J'^2 ) gam
 *    sqrt(gam^2-1) Exp(-gam/tdl) Exp(1/tdl)    ------ (eq 1)
 * Large tdl:
 *   (1/beta) ( coeffA^2 J^2 + coeffB^2 J'^2 ) gam
 *    sqrt(gam^2-1) Exp(-gam/tdl)               ------ (eq 2)
 *
 * Depending on the value of K2_EQ, either (eq 1) or (eq 2) is returned.
 *
 * Note that coeffA and coeffB are not the same as those in the total 
 *  emissivity. Here they include the index of refraction and polarization
 *  coefficients (Ramaty 1969, Eq.(1),(2)).
 */
static double jnuOX_integrand_modified(int flag, double gamma, double tdl, 
                                       double n, double nuratio, double cos_theta, 
                                       double sin_theta, double B, double nrhoe)
{
        double jn, d_jn, arg_jn, n_0, beta, mu, sinxi, tmp, tmp1, tmp2;
        double T, L;

        if(gamma==1.) return(0.); /* if electron is at rest, there's no emission */

        if(flag!=0) return_T_L(&T, &L, flag, nuratio, nrhoe, B, cos_theta, sin_theta);

        n_0 = nuratio * gamma ;

        if(gamma < 1.) {
                fprintf(stderr,"ERROR in jnuOX_integrand_modified(), gamma < 1\n");
                fprintf(stderr,"%.16e %.16e %.16e %.16e %.16e %.16e\n",
                               gamma,tdl,n,nuratio,cos_theta,sin_theta);
                exit(-1);
        }
        beta = sqrt( (gamma-1.)*(gamma+1.) ) / gamma;

        mu = ( n_0 - n ) / ( n_0 * beta * cos_theta );

        if(mu >= -1. && mu <= 1.) {
                sinxi = sqrt( 1. - mu*mu );
        } else {
#if FLAG_ERR2 == ERR2_ON_SCREEN_EXIT
                fprintf(stderr,"ERROR in jnuOX_integrand_modified(), ");
                fprintf(stderr,"\\mu (i.e., cos(\\xi)) out of bound\n");
                fprintf(stderr," mu,n_0,n,beta,cos_theta = ");
                fprintf(stderr,"%.16e %.16e %.16e %.16e %.16e\n",mu,n_0,n,beta,cos_theta);
                exit(-1);
#elif FLAG_ERR2 == ERR2_ON_SCREEN_CONT
                fprintf(stderr,"ERROR in jnuOX_integrand_modified(), ");
                fprintf(stderr,"\\mu (i.e., cos(\\xi)) out of bound\n");
                fprintf(stderr," mu,n_0,n,beta,cos_theta = ");
                fprintf(stderr,"%.16e %.16e %.16e %.16e %.16e\n",mu,n_0,n,beta,cos_theta);
                if(mu < -1.) {
                        mu = -1.;
                        sinxi = 0.;
                } else {
                        mu = 1.;
                        sinxi = 0.;
                }
#elif FLAG_ERR2 == ERR2_TO_FILE_CONT
                fprintf(errorfixedpt,"ERROR in jnuOX_integrand_modified(), ");
                fprintf(errorfixedpt,"\\mu (i.e., cos(\\xi)) out of bound\n");
                fprintf(errorfixedpt," mu,n_0,n,beta,cos_theta = ");
                fprintf(errorfixedpt,"%.16e %.16e %.16e %.16e %.16e\n",
                                     mu,n_0,n,beta,cos_theta);
                fflush(errorfixedpt);
                if(mu < -1.) {
                        mu = -1.;
                        sinxi = 0.;
                } else {
                        mu = 1.;
                        sinxi = 0.;
                }
#elif FLAG_ERR2 == ERR2_FIX_CONT
                if(mu < -1.) {
                        mu = -1.;
                        sinxi = 0.;
                } else {
                        mu = 1.;
                        sinxi = 0.;
                }
#elif FLAG_ERR2 == ERR2_IGNORE
#else
  #error "FLAG_ERR2 value unknown"
#endif
        }

        /* Ramaty 1969 Eq.(8) (seems to be wrong, see Chanmugam & Dulk 81 Eq.(3) */
        //arg_jn = n_0 * n_refrac * beta * sin_theta * sinxi
        //         / (1. - n_refrac * beta * cos_theta * mu);
        arg_jn = n_0 * beta * sin_theta * sinxi;

        /* Bessel function routine: */
        d_jn = my_Bessel_dJ(n, arg_jn, &jn);

        if(flag==0) {
                tmp1 = jn * (cos_theta - beta * mu) / sin_theta;
                tmp2 = d_jn * beta * sinxi;

                tmp = tmp1*tmp1 + tmp2*tmp2;
        } else {
                tmp1 = T * jn * (cos_theta - beta * mu) / sin_theta;
                tmp2 = d_jn * beta * sinxi;

                /* Ramaty 1969 Eq.(7), second line (without square) */
                tmp = tmp1 + tmp2;
                /* Ramaty 1969 Eq.(7), first line (after summation sign), and second line */
                tmp = tmp * tmp / (1. + T*T);
        }

        DISTRIBUTION_PARAMETER;
        return(tmp*ne(gamma,mu,&ne_para[0])/beta);
}

/*============================================================================*/

static double jnu_Summation(void *par_n, int nS1, int nS2)
{
        TYPE_PAR_n *p = (TYPE_PAR_n *)par_n;
        double ans = 0.;
        int i;

        /* Initialize gMaxAbsErr as the smallest possible number.
         * It will be replaced by the maximum absolute error of the g integration */
        p->gMaxAbsErr = -DBL_MAX;

        if(nS1>nS2) return(0.);

        for(i=nS1;i<=nS2;i++) {
                ans += func_n(i,par_n);
        }

        return(ans);
}

/*----------------------------------------------------------------------------*/

static double jnu_Integration(void *par_jnu_I, double nI1, double nI2)
{
        TYPE_PAR_jnu_I *p = (TYPE_PAR_jnu_I *)par_jnu_I;
        double ans;
        int status;
        TYPE_PAR_n *pn = (TYPE_PAR_n *)(p->integrand_n).params;

        /* Initialize gMaxAbsErr as the smallest possible number.
         * It will be replaced by the maximum absolute error of the g integration */
        pn->gMaxAbsErr = -DBL_MAX;

#if INTEGRATOR == GSL_QAG
        status = gsl_integration_qag(&(p->integrand_n), nI1, nI2,
                                     p->nAbsEps, p->nRelEps,
                                     p->n_wksp_n, GSL_QAG_RULE_n, p->wksp_n_pt,
                                     &ans, &(p->nAbsErr));
#elif INTEGRATOR == GSL_QNG
        size_t neval;
        status = gsl_integration_qng(&(p->integrand_n), nI1, nI2,
                                     p->nAbsEps, p->nRelEps, &ans, 
                                     &(p->nAbsErr), &neval );
#else
  #error "INTEGRATOR is not set correctly"
#endif

        if(status) {
#if FLAG_ERR1 == ERR1_ON_SCREEN_EXIT
                fprintf(stderr,"GSL error in  jnu_Integration().  gsl_errno = %d\n",status);
                exit(-1);
#elif FLAG_ERR1 == ERR1_ON_SCREEN_CONT
                fprintf(stderr,"GSL error in  jnu_Integration().  gsl_errno = %d\n",status);
                fprintf(stderr,"gsl_errno = %d\n",status);
#elif FLAG_ERR1 == ERR1_TO_FILE_CONT
                fprintf(errorpt,"GSL error in  jnu_Integration().  gsl_errno = %d\n",status);
                fprintf(errorpt,"gBegin,gEnd,nuratio = %.16e %.16e %.16e\n",
                                pn->gBegin, pn->gEnd, (pn->par_g_pt)->nuratio);
                fprintf(errorpt,"tdl,cos_theta,sin_theta = %.16e %.16e %.16e\n",
                                (pn->par_g_pt)->tdl, (pn->par_g_pt)->cos_theta,
                                (pn->par_g_pt)->sin_theta);
                fprintf(errorpt,"n,ans,nAbsEps, = %.16e %.16e %.16e\n",
                                (pn->par_g_pt)->n, ans, p->nAbsEps);
                fprintf(errorpt,"nRelEps,nAbsErr,gAbsEps, = %.16e %.16e %.16e\n",
                                p->nRelEps, p->nAbsErr, pn->gAbsEps);
                fprintf(errorpt,"gRelEps,gMaxAbsErr = %.16e %.16e\n",
                                pn->gRelEps, pn->gMaxAbsErr);
                fflush(errorpt);
#elif FLAG_ERR1 == ERR1_IGNORE
#else
  #error "FLAG_ERR1 value unknown"
#endif
        }

        return(ans);
}

/*============================================================================*/
/* do summation for small n and integration for large n */
double summation_and_integration(void *par_n, void *par_Integration)
{
        double ans;
        double nBegin=0.,nEnd=0.,nBeginI=0.,nEndI=0.,ntmp;
        int nBeginS=0,nEndS=0;
        double ansI,ansS;
        int flagI,flagS;

        TYPE_PAR_n *p = (TYPE_PAR_n *)par_n;

#if (ELECTRON_DIST == THERMAL)
        nBegin = (p->par_g_pt)->nuratio * fabs((p->par_g_pt)->sin_theta);
        double gamTmp;
        gamTmp = (p->par_g_pt)->tdl + 1.;
        /* Delta n = gamma x ( 1 - beta^2 cos_theta^2)
         * For small x (x < TDL^2), gamma = TDL + 1;
         * for large x (x > TDL^2), gamma = pow(2 x TDL,1/3) */
        nEnd = COEFF_DELTA_N * (gamTmp + pow(2. * (p->par_g_pt)->nuratio * (p->par_g_pt)->tdl , 1./3.))
                 * (p->par_g_pt)->nuratio * (1. - (p->par_g_pt)->cos_theta * (p->par_g_pt)->cos_theta / (gamTmp * gamTmp));
        nEnd += nBegin;
#elif (ELECTRON_DIST == NONTHERMAL \
    || ELECTRON_DIST == NONTHERMAL_ANISO \
    || ELECTRON_DIST == GAUSSIAN)
        nBegin = (p->par_g_pt)->nuratio * fabs((p->par_g_pt)->sin_theta);
        nEnd = nBegin + nEND;
#else
  #error Shouldn't arrive here
#endif

        flagS = 1; /* turn on summation in general */
        flagI = 1; /* turn on integration in general */
        if(nBegin <= 1.) {
                nBegin = 1.; /* lowest "hard limit" is 1 */
                nBeginS = 1;
        } else if (nBegin <= n_INT_MIN) {
                nBeginS = (int)ceil(nBegin); /* smallest integer within the boundary */
        } else {
                flagS = 0; /* turn off summation */
        }
        if(nEnd <= n_INT_MIN) {
                flagI = 0; /* turn off integration */
                nEndS = (int)nEnd; /* largest integer within the boundary */
        } else {
                nEndS = (int)n_INT_MIN;
                nBeginI = nEndS + 0.5;
                nEndI = nEnd;
                //if(nBeginI >= nEndI) flagI = 0; /* turn off integration */
                if(nBeginI >= nEndI) nEndI = nBeginI + 10. ;
        }

        /* sanity check */
        if(flagS==0 && flagI==0) {
                fprintf(stderr,"ERR: both summation and integration are turned off!\n");
                exit(92);
        }

        /* summation for small n */
        if(flagS) {
                ansS = jnu_Summation(&(*par_n), nBeginS, nEndS);
        } else {
                ansS = 0.;
        }
        if(flagI) {
#if INTEGRATE_WITH_LOGN == TRUE
                ansI = jnu_Integration(&(*par_Integration), log(nBeginI), log(nEndI));
#elif INTEGRATE_WITH_LOGN == FALSE
                ansI = jnu_Integration(&(*par_Integration), nBeginI, nEndI);
#endif
                nBeginI = nEndI; nEndI *= 2.;
        } else {
                ansI = 0.;
                nBeginI = nEndS + 0.5; nEndI = nEndS * 2.;
        }
        ans = ansS + ansI;
        flagS = 0; /* make sure summation is turned off */
        flagI = 1; /* make sure integration is turned on */

#if (ADAPTIVE == TRUE)
        double ansLIST[NMAX_ADAPTIVE],total;
        int iIT;
        total = ans;
        ansLIST[0] = ans;
        for(iIT=1;iIT<NMAX_ADAPTIVE;iIT++) {
  #if INTEGRATE_WITH_LOGN == TRUE
                ans = jnu_Integration(&(*par_Integration), log(nBeginI), log(nEndI));
  #elif INTEGRATE_WITH_LOGN == FALSE
                ans = jnu_Integration(&(*par_Integration), nBeginI, nEndI);
  #endif
                ansLIST[iIT] = ans;
                total += ans;
                if(total != 0. && (ansLIST[iIT]==0. || fabs(ansLIST[iIT])<EPS_ADAPTIVE*fabs(total))) break;
                nBeginI = nEndI; nEndI *= 2.;
        }
        ans = total;
#endif

        return(ans);
}

/*----------------------------------------------------------------------------*/

/* assign parameters to pointer */
void assign_par(TYPE_PAR_n *par_n, TYPE_PAR_jnu_I *par_Integration, 
                double (*integrand_g_fun)(double x,void*P),
                double nuratio, double nrhoe, double tdl, double B,
                double sin_theta, double cos_theta,
                size_t n_wksp,
                gsl_integration_workspace *wksp_g,
                gsl_integration_workspace *wksp_n,
                int flag_Tx_interp, void *Tx_list)
{
        gsl_function integrand_g;
        TYPE_PAR_g par_g;

        par_g.cos_theta = cos_theta;
        par_g.sin_theta = sin_theta;
        par_g.nuratio   = nuratio;
        par_g.tdl       = tdl;
        par_g.B         = B;
        par_g.nrhoe     = nrhoe;
        par_g.flag_Tx_interp = flag_Tx_interp;
        par_g.Tx_list = Tx_list;
        integrand_g.function = &(*integrand_g_fun);
        integrand_g.params   = &par_g;

        (*par_n).par_g_pt = &par_g;
        (*par_n).integrand_g_pt = &integrand_g;
        (*par_n).gAbsEps = EPS_ABS_g;
        (*par_n).gRelEps = EPS_REL_g;
        (*par_n).n_wksp_g = n_wksp;
        (*par_n).wksp_g_pt = wksp_g;

        (*par_Integration).integrand_n.function = &func_n;
        (*par_Integration).integrand_n.params   = &(*par_n);
        (*par_Integration).nAbsEps = EPS_ABS_n;
        (*par_Integration).nRelEps = EPS_REL_n;
        (*par_Integration).n_wksp_n = n_wksp;
        (*par_Integration).wksp_n_pt = wksp_n;
}

/*============================================================================*/
/* Specific emissivity calculated by using Gamma-n integration */
/* Can be one of ordinary mode (O), extraordinary mode (X),
 * Stokes parameters I, Q, V, or total */
double return_jnu(int flagMODE,
                 double nu, double rho, double Te, double B, double theta,
                 size_t n_wksp, gsl_integration_workspace *wksp_g,
                 gsl_integration_workspace *wksp_n)
{
        double nrhoe,nuratio,nu_c,sin_theta,cos_theta,tdl,tdlinv;
        TYPE_PAR_n par_n;
        TYPE_PAR_jnu_I par_I;
        void *list = NULL;

        nrhoe = rho/(C_me+C_mp);
        sin_theta = sin(theta);
        cos_theta = cos(theta);
        nu_c = C_e*B/(C_2pi*C_me*C_c);
        nuratio = nu/nu_c;
        tdl = C_k * Te / C_me_c2;
        tdlinv = 1./tdl;

        /* assign parameters for integration */
        switch(flagMODE) {
                case TOTAL:
                        assign_par(&par_n,&par_I,func_g,
                                   nuratio,nrhoe,tdl,B,sin_theta,cos_theta,
                                   n_wksp,wksp_g,wksp_n,0,list);
                        break;
                case MODE_O:
                        assign_par(&par_n,&par_I,funcO_g,
                                   nuratio,nrhoe,tdl,B,sin_theta,cos_theta,
                                   n_wksp,wksp_g,wksp_n,0,list);
                        break;
                case MODE_X:
                        assign_par(&par_n,&par_I,funcX_g,
                                   nuratio,nrhoe,tdl,B,sin_theta,cos_theta,
                                   n_wksp,wksp_g,wksp_n,0,list);
                        break;
                case STOKES_I:
                        assign_par(&par_n,&par_I,funcI_g,
                                   nuratio,nrhoe,tdl,B,sin_theta,cos_theta,
                                   n_wksp,wksp_g,wksp_n,0,list);
                        break;
                case STOKES_Q:
                        assign_par(&par_n,&par_I,funcQ_g,
                                   nuratio,nrhoe,tdl,B,sin_theta,cos_theta,
                                   n_wksp,wksp_g,wksp_n,0,list);
                        break;
                case STOKES_V:
                        assign_par(&par_n,&par_I,funcV_g,
                                   nuratio,nrhoe,tdl,B,sin_theta,cos_theta,
                                   n_wksp,wksp_g,wksp_n,0,list);
                        break;
                default:
                        fprintf(stderr,"ERR: invalid value of flagMODE in alphanu()\n");
                        fprintf(stderr,"     flagMODE = %d\n",flagMODE);
                        exit(91);
                        break;
        }

        /* parameters used for calculating coefficient */
        double para[] = {nrhoe,nu,cos_theta};

        /* do the summation (for low n) and integration */
        return(summation_and_integration(&par_n,&par_I)
               * extra_factor_modified(&para[0]));
}

/*----------------------------------------------------------------------------*/

/* Wrapper of return_jnu()
 * If flag EXTRAPOLATE is TRUE, and if theta is close to pi/2, then do
 *  extrapolation to get result. */
double jnu(int flagMODE,
                 double nu, double rho, double Te, double B, double theta,
                 size_t n_wksp, gsl_integration_workspace *wksp_g,
                 gsl_integration_workspace *wksp_n)
{
#if (EXTRAPOLATE == FALSE)
        /* no extrapolation */
        return(return_jnu(flagMODE,nu,rho,Te,B,theta,n_wksp,wksp_g,wksp_n));
#elif (EXTRAPOLATE == TRUE)
        /* extrapolate near to Pi/2 */

        /* If theta is in (Pi/2 - DELTA_THETA , Pi/2 + DELTA_THETA), then 
         *  extrapolate.  Note that DELTA_THETA is assumed to be positive.  
         *  It should also be small (say, less than 0.0017453 rad or 
         *  0.1 deg). */
        /* Results are calculated at (Pi/2 + sign * DELTA_THETA) and
         *   (Pi/2 + sign * 2 * DELTA_THETA), and then extrapolate to theta.
         *   sign is -1 if theta < Pi/2, and +1 if theta > Pi/2 */

        int sign;
        double dtheta = theta - C_pi_2;
        if(dtheta < 0.) sign=-1;
        else sign=1;

        /* check whether we are close to pi/2 */
        if( fabs(dtheta) > DELTA_THETA ) {
                /* no extrapolation */
                return(return_jnu(flagMODE,nu,rho,Te,B,theta,n_wksp,wksp_g,wksp_n));
        } else {
                /* extrapolation */
		double theta1 = C_pi_2 + sign * DELTA_THETA;
                double theta2 = C_pi_2 + sign * 2. * DELTA_THETA;
                double ans1 = return_jnu(flagMODE,nu,rho,Te,B,theta1,n_wksp,wksp_g,wksp_n);
                double ans2 = return_jnu(flagMODE,nu,rho,Te,B,theta2,n_wksp,wksp_g,wksp_n);
                return( (theta-theta1)*(ans2-ans1)/(theta2-theta1) + ans1 );
        }
#else
  #error not valid option for EXTRAPOLATE
#endif
}

/*============================================================================*/

/* alphanu for total (both O and X modes) */
static double alphanufunc_g(double g, void *par)
{
        TYPE_PAR_g *p = (TYPE_PAR_g *)par;

        return(alphanu_integrand(g,p->tdl,p->n,p->nuratio,p->cos_theta,
                                 p->sin_theta,p->B,p->nrhoe));
}

/*----------------------------------------------------------------------------*/

/* alphanu for ordinary (O) mode */
static double alphanufuncO_g(double g, void *par)
{
        TYPE_PAR_g *p = (TYPE_PAR_g *)par;

        return(alphanuOX_integrand(1,g,p->tdl,p->n,p->nuratio,p->cos_theta,
                                   p->sin_theta,p->B,p->nrhoe,p->flag_Tx_interp,p->Tx_list));
}

/*----------------------------------------------------------------------------*/

/* alphanu for extraordinary (X) mode */
static double alphanufuncX_g(double g, void *par)
{
        TYPE_PAR_g *p = (TYPE_PAR_g *)par;

        return(alphanuOX_integrand(-1,g,p->tdl,p->n,p->nuratio,p->cos_theta,
                                   p->sin_theta,p->B,p->nrhoe,p->flag_Tx_interp,p->Tx_list));
}

/*----------------------------------------------------------------------------*/

/* alphanu for Stokes I */
static double alphanufuncI_g(double g, void *par)
{
        TYPE_PAR_g *p = (TYPE_PAR_g *)par;

        return(alphanuStokes_integrand(1,g,p->tdl,p->n,p->nuratio,p->cos_theta,
                                       p->sin_theta,p->B,p->nrhoe));
}

/*----------------------------------------------------------------------------*/

/* alphanu for Stokes Q */
static double alphanufuncQ_g(double g, void *par)
{
        TYPE_PAR_g *p = (TYPE_PAR_g *)par;

        return(alphanuStokes_integrand(2,g,p->tdl,p->n,p->nuratio,p->cos_theta,
                                       p->sin_theta,p->B,p->nrhoe));
}

/*----------------------------------------------------------------------------*/

/* alphanu for Stokes V */
static double alphanufuncV_g(double g, void *par)
{
        TYPE_PAR_g *p = (TYPE_PAR_g *)par;

        return(alphanuStokes_integrand(4,g,p->tdl,p->n,p->nuratio,p->cos_theta,
                                       p->sin_theta,p->B,p->nrhoe));
}

/*----------------------------------------------------------------------------*/

/* Dimensional numbers and coefficients in front of the result of integration.
 * Absorption Coeff = (this factor) * (result of n & g integrations) */
static double extra_factor_alphanu(int flagMODE, double nrhoe, double nu, double cos_theta)
{
        double cst;
        switch(flagMODE) {
                case TOTAL:
                case STOKES_I: 
                case STOKES_Q: 
                case STOKES_V: 
                        cst = 0.5;
                        break;
                case MODE_O:
                case MODE_X:
                        cst = 1.;
                        break;
                default:
                        fprintf(stderr,"ERR: invalid value of flagMODE\n");
                        exit(90);
                        break;
        }
        return(-nrhoe*C_2pi*C_c*C_e2/(nu*C_me*fabs(cos_theta)) * cst);
}

/*============================================================================*/
/* Distribution function n_e(\gamma,\mu), without number density nrhoe
 * gamma is Lorentz factor, and mu is cos(theta) */
/* In other words, n_e is normalized:
 * Integrate[ n_e(\gamma,\nu} , {\gamma,1,Infinity} , {\mu,-1,1} ] = 1 */
static double ne(double gamma, double mu, double *ne_para)
{
#if (ELECTRON_DIST == THERMAL)
        double tdl = ne_para[0];
        double tdlinv = 1./tdl;
        if(gamma < 1.) {
                fprintf(stderr,"ERROR in ne(), gamma < 1\n");
                fprintf(stderr,"gamma, mu, tdl = %.16e %.16e %.16e\n",
                                gamma, mu, tdl);
                exit(-1);
        }
#elif (ELECTRON_DIST == NONTHERMAL)
        double index,gam1,gam2;
        index = ne_para[0];
        gam1 = ne_para[1];
        gam2 = ne_para[2];
        if(gamma < 1.) {
                fprintf(stderr,"ERROR in ne(), gamma < 1\n");
                fprintf(stderr,"gamma, mu, index, gam1, gam2 = %.16e %.16e %.16e %.16e %.16e\n",
                                gamma, mu, index, gam1, gam2);
          exit(-1);
        }
#elif (ELECTRON_DIST == NONTHERMAL_ANISO)
        double index,gam1,gam2,xic;
        index = ne_para[0];
        gam1 = ne_para[1];
        gam2 = ne_para[2];
        xic = ne_para[3];
        if(gamma < 1.) {
                fprintf(stderr,"ERROR in ne(), gamma < 1\n");
                fprintf(stderr,"gamma, mu, index, gam1, gam2, xic = %.16e %.16e %.16e %.16e %.16e %.16e\n",
                                gamma, mu, index, gam1, gam2, xic);
          exit(-1);
        }
#elif (ELECTRON_DIST == GAUSSIAN)
        double gam0, delta;
        gam0 = ne_para[0];
        delta = ne_para[1];
        if(gamma < 1.) {
                fprintf(stderr,"ERROR in ne(), gamma < 1\n");
                fprintf(stderr,"gam0, delta = %.16e %.16e\n",
                                gam0, delta);
                exit(-1);
        }
#endif

#if (ELECTRON_DIST == THERMAL)
        return( exp_factor( (0.5*gamma*sqrt(gamma*gamma-1.)*tdlinv*
                            K2_tdlinv_inv_new(tdlinv)) , (-gamma*tdlinv) ) ) ;
#elif (ELECTRON_DIST == NONTHERMAL)
        return( gamma < gam1 ? 0. : gamma > gam2 ? 0. :
                0.5 * pow( gamma , -index ) / COEFF_INT );
#elif (ELECTRON_DIST == NONTHERMAL_ANISO)
        double xi = acos(mu);
        double f2 = (xi<xic ?
                     pow(sin(C_pi_2*xi/xic),6) :
                     xi>C_pi-xic ? pow(sin(C_pi_2*(xi-C_pi)/xic),6) : 1.);
        return( gamma < gam1 ? 0. : gamma > gam2 ? 0. :
                (pow( gamma*gamma-1. , (1.-index)/2. )
              / COEFF_G_INT) * f2/COEFF_MU_INT );
#elif (ELECTRON_DIST == GAUSSIAN)
        return( 0.5 * exp_factor( 1./(delta*sqrt(C_pi)) ,
               -(gamma-gam0)*(gamma-gam0)/(delta*delta) ) );
#endif
}

/*----------------------------------------------------------------------------*/
/* wrappers for gsl */
typedef struct {
        double gamma;
        double mu;
        double tdl;
} PAR_ne;
/*----------------------------------------*/
double ne_gsl_dg(double gamma, void *params)
{
        PAR_ne *p = (PAR_ne *)params;
        if(gamma < 1.) {
                fprintf(stderr,"ERROR in ne_gsl_dg(), gamma < 1\n");
                fprintf(stderr,"gamma, p->mu, p->tdl = %.16e %.16e %.16e\n",
                                gamma, p->mu, p->tdl);
                exit(-1);
        }
        double beta = sqrt( (gamma-1.)*(gamma+1.) ) / gamma;
        double tdl = p->tdl;

        DISTRIBUTION_PARAMETER;
        return(ne(gamma,p->mu,&ne_para[0])/(gamma*gamma*beta));
}
/*----------------------------------------*/
double ne_gsl_dm(double mu, void *params)
{
        PAR_ne *p = (PAR_ne *)params;
        if(p->gamma < 1.) {
                fprintf(stderr,"ERROR in ne_gsl_dg(), gamma < 1\n");
                fprintf(stderr,"p->gamma, mu, p->tdl = %.16e %.16e %.16e\n",
                                p->gamma, mu, p->tdl);
                exit(-1);
        }
        double beta = sqrt( (p->gamma-1.)*(p->gamma+1.) ) / p->gamma;
        double tdl = p->tdl;

        DISTRIBUTION_PARAMETER;
        return(ne(p->gamma,mu,&ne_para[0]));
}

/*============================================================================*/
/* -- integrand (without constant factors) of the gamma integration of alphanu, 
 * for a certain n. Include both O and X modes.
 */
/* Ramaty 1969 ApJ, 158, 753, Eq.(14), with Eq.(7), Eq.(1) and Eq.(2).
 * Also see Chanmugam & Dulk 1981 ApJ, 244, 569, Eq.(5), which does the 
 *  integration in p and \phi instead of \gamma and \phi.
 * Also see Stupp 2000 MNRAS, 311, 251, Eq.(12), which does not contain the
 *  factor 1+d(log(n_ref))/d(log(\nu)), and contains the longitudinal 
 *  contribution, and keeps a pitch angle dependence.
 */
static double alphanu_integrand(double gamma, double tdl, 
                                double n, double nuratio, double cos_theta, 
                                double sin_theta, double B, double nrhoe)
{
 return(alphanuStokes_integrand(1,gamma,tdl,n,nuratio,cos_theta,
                                sin_theta,B,nrhoe));
}

/*----------------------------------------------------------------------------*/
/* -- integrand (without constant factors) of the gamma integration of alphanu, 
 * for a certain n. Only one of the two modes.
 * Ordinary (O) mode      -->  flag = +1
 * Extrardinary (X) mode  -->  flag = -1
 */
/* Ramaty 1969 ApJ, 158, 753, Eq.(14), with Eq.(7), Eq.(1) and Eq.(2).
 * Also see Chanmugam & Dulk 1981 ApJ, 244, 569, Eq.(5), which does the 
 *  integration in p and \phi instead of \gamma and \phi.
 * Also see Stupp 2000 MNRAS, 311, 251, Eq.(12), which does not contain the
 *  factor 1+d(log(n_ref))/d(log(\nu)), and contains the longitudinal 
 *  contribution, and keeps a pitch angle dependence.
 */
/*
 * Small tdl:
 *   (1/beta) ( coeffA^2 J^2 + coeffB^2 J'^2 ) gam
 *    sqrt(gam^2-1) Exp(-gam/tdl) Exp(1/tdl)    ------ (eq 1)
 * Large tdl:
 *   (1/beta) ( coeffA^2 J^2 + coeffB^2 J'^2 ) gam
 *    sqrt(gam^2-1) Exp(-gam/tdl)               ------ (eq 2)
 *
 * Depending on the value of K2_EQ, either (eq 1) or (eq 2) is returned.
 *
 * Note that coeffA and coeffB are not the same as those in the total 
 *  emissivity. Here they include the index of refraction and polarization
 *  coefficients (Ramaty 1969, Eq.(1),(2)).
 */
static double alphanuOX_integrand(int flag, double gamma, double tdl, 
                                  double n, double nuratio, double cos_theta, 
                                  double sin_theta, double B, double nrhoe,
                                  int flag_Tx_interp, void *Tx_list)
{
        double jn, d_jn, arg_jn, n_0, beta, mu, sinxi, tmp, tmp1, tmp2;
        double T, L, Df;
        double d_ne_gam, d_ne_mu;
        double coeffDgam, coeffDmu;
        double abserr_g,abserr_m;
        gsl_function Fg,Fm;
        PAR_ne par;

        if(gamma==1.) return(0.); /* if electron is at rest, there are no emission
                                   * and no absorption */

        if(flag_Tx_interp == 0) {
                return_T_L(&T, &L, -1, nuratio, nrhoe, B, cos_theta, sin_theta);
        } else if(flag_Tx_interp == 1) {
                /* find Tx by interpolation of data from alphanu_Q,V calculation */
                TYPE_Tx *p = (TYPE_Tx *)Tx_list;
                T = Tx_interp(p, nuratio * p->nuc);
        } else {
                /* shouldn't arrive here */
                fprintf(stderr,"ERROR in alphanuOX_integrand(), ");
                fprintf(stderr,"flag_Tx_interp = %d has invalid value\n",flag_Tx_interp);
                exit(-1);
        }

        n_0 = nuratio * gamma ;

        if(gamma < 1.) {
                fprintf(stderr,"ERROR in alphanuOX_integrand(), gamma < 1\n");
                fprintf(stderr,"%.16e %.16e %.16e %.16e %.16e %.16e\n",
                               gamma,tdl,n,nuratio,cos_theta,sin_theta);
                exit(-1);
        }
        beta = sqrt( (gamma-1.)*(gamma+1.) ) / gamma;

        mu = ( n_0 - n ) / ( n_0 * beta * cos_theta );

        if(mu >= -1. && mu <= 1.) {
                sinxi = sqrt( 1. - mu*mu );
        } else {
#if FLAG_ERR2 == ERR2_ON_SCREEN_EXIT
                fprintf(stderr,"ERROR in alphanuOX_integrand(), ");
                fprintf(stderr,"\\mu (i.e., cos(\\xi)) out of bound\n");
                fprintf(stderr," mu,n_0,n,beta,cos_theta = ");
                fprintf(stderr,"%.16e %.16e %.16e %.16e %.16e\n",mu,n_0,n,beta,cos_theta);
                exit(-1);
#elif FLAG_ERR2 == ERR2_ON_SCREEN_CONT
                fprintf(stderr,"ERROR in alphanuOX_integrand(), ");
                fprintf(stderr,"\\mu (i.e., cos(\\xi)) out of bound\n");
                fprintf(stderr," mu,n_0,n,beta,cos_theta = ");
                fprintf(stderr,"%.16e %.16e %.16e %.16e %.16e\n",mu,n_0,n,beta,cos_theta);
                if(mu < -1.) {
                        mu = -1.;
                        sinxi = 0.;
                } else {
                        mu = 1.;
                        sinxi = 0.;
                }
#elif FLAG_ERR2 == ERR2_TO_FILE_CONT
                fprintf(errorfixedpt,"ERROR in alphanuOX_integrand(), ");
                fprintf(errorfixedpt," \\mu (i.e., cos(\\xi)) out of bound\n");
                fprintf(errorfixedpt," mu,n_0,n,beta,cos_theta = ");
                fprintf(errorfixedpt,"%.16e %.16e %.16e %.16e %.16e\n",
                                     mu,n_0,n,beta,cos_theta);
                fflush(errorfixedpt);
                if(mu < -1.) {
                        mu = -1.;
                        sinxi = 0.;
                } else {
                        mu = 1.;
                        sinxi = 0.;
                }
#elif FLAG_ERR2 == ERR2_FIX_CONT
                if(mu < -1.) {
                        mu = -1.;
                        sinxi = 0.;
                } else {
                        mu = 1.;
                        sinxi = 0.;
                }
#elif FLAG_ERR2 == ERR2_IGNORE
#else
  #error "FLAG_ERR2 value unknown"
#endif
        }

  /* Ramaty 1969 Eq.(8) (seems to be wrong, see Chanmugam & Dulk 81 Eq.(3) */
        arg_jn = n_0 * beta * sin_theta * sinxi;

  /* Scott's Bessel function routine: */
        d_jn = my_Bessel_dJ(n, arg_jn, &jn);

        if(flag == 1) {
                tmp1 =  jn   * (cos_theta - beta * mu) / sin_theta;
                tmp2 = -d_jn * T * beta * sinxi;
        } else if(flag == -1) {
                tmp1 = jn   * T * (cos_theta - beta * mu) / sin_theta;
                tmp2 = d_jn * beta * sinxi;
        } else { /* shouldn't have arrive here */
                fprintf(stderr,"ERROR in alphanuOX_integrand(), flag=%d is not allowed\n",flag);
                exit(1);
        }

        /* Ramaty 1969 Eq.(7), second line (without square) */
        tmp = tmp1 + tmp2;
        /* Ramaty 1969 Eq.(7), first line (after summation sign), and second line */
        tmp = tmp * tmp / (1. + T*T);

        coeffDgam = gamma*gamma/C_c2; // coefficient of d/d\gamma term
        coeffDmu = (beta*cos_theta - mu) / (gamma*beta*beta*beta*C_c2); // coefficient of d/d\xi

        par.gamma = gamma;
        par.mu    = mu;
        par.tdl   = tdl;
        Fg.function = &ne_gsl_dg;
        Fm.function = &ne_gsl_dm;
        Fg.params = &par;
        Fm.params = &par;
        if(gamma-DGAMMA<1.) {  /* use forward method */
                gsl_deriv_forward(&Fg, gamma, DGAMMA, &d_ne_gam, &abserr_g);
        } else {               /* use central method */
                gsl_deriv_central(&Fg, gamma, DGAMMA, &d_ne_gam, &abserr_g);
        }
        if(mu-DMU<-1.) {       /* use forward method */
                gsl_deriv_forward(&Fm, mu, DMU, &d_ne_mu, &abserr_m);
        } else if(mu+DMU>1.) { /* use backward method */
                gsl_deriv_backward(&Fm, mu, DMU, &d_ne_mu, &abserr_m);
        } else {               /* use central method */
                gsl_deriv_central(&Fm, mu, DMU, &d_ne_mu, &abserr_m);
        }
        Df = coeffDgam * d_ne_gam + coeffDmu * d_ne_mu;

        return(tmp*Df);
}

/*----------------------------------------------------------------------------*/
/* -- integrand (without constant factors) of the gamma integration of alphanu, 
 * for a certain n. Only one of Stokes I, Q, V.
 * Stokes I  -->  flag = 1
 * Stokes Q  -->  flag = 2
 * Stokes V  -->  flag = 4
 */
/* Ramaty 1969 ApJ, 158, 753, Eq.(14), with Eq.(7), Eq.(1) and Eq.(2).
 * Also see Chanmugam & Dulk 1981 ApJ, 244, 569, Eq.(5), which does the 
 *  integration in p and \phi instead of \gamma and \phi.
 * Also see Stupp 2000 MNRAS, 311, 251, Eq.(12), which does not contain the
 *  factor 1+d(log(n_ref))/d(log(\nu)), and contains the longitudinal 
 *  contribution, and keeps a pitch angle dependence.
 */
/*
 * Small tdl:
 *   (1/beta) ( coeffA^2 J^2 + coeffB^2 J'^2 ) gam
 *    sqrt(gam^2-1) Exp(-gam/tdl) Exp(1/tdl)    ------ (eq 1)
 * Large tdl:
 *   (1/beta) ( coeffA^2 J^2 + coeffB^2 J'^2 ) gam
 *    sqrt(gam^2-1) Exp(-gam/tdl)               ------ (eq 2)
 *
 * Depending on the value of K2_EQ, either (eq 1) or (eq 2) is returned.
 *
 * Note that coeffA and coeffB are not the same as those in the total 
 *  emissivity. Here they include the index of refraction and polarization
 *  coefficients (Ramaty 1969, Eq.(1),(2)).
 */
static double alphanuStokes_integrand(int flag, double gamma, double tdl, 
                                      double n, double nuratio, double cos_theta, 
                                      double sin_theta, double B, double nrhoe)
{
      double jn, d_jn, arg_jn, n_0, beta, mu, sinxi, tmp, tmp1, tmp2;
      double Df;
      double d_ne_gam, d_ne_mu;
      double coeffDgam, coeffDmu;
      double abserr_g,abserr_m;
      gsl_function Fg,Fm;
      PAR_ne par;

      if(gamma==1.) return(0.); /* if electron is at rest, there are no emission
                                 * and no absorption */

      n_0 = nuratio * gamma ;

      if(gamma < 1.) {
            fprintf(stderr,"ERROR in alphanuStokes_integrand(), gamma < 1\n");
            fprintf(stderr,"%.16e %.16e %.16e %.16e %.16e %.16e\n",
                           gamma,tdl,n,nuratio,cos_theta,sin_theta);
            exit(-1);
      }
      beta = sqrt( (gamma-1.)*(gamma+1.) ) / gamma;

      mu = ( n_0 - n ) / ( n_0 * beta * cos_theta );

      if(mu >= -1. && mu <= 1.) {
            sinxi = sqrt( 1. - mu*mu );
      } else {
#if FLAG_ERR2 == ERR2_ON_SCREEN_EXIT
            fprintf(stderr,"ERROR in alphanuStokes_integrand(), ");
            fprintf(stderr,"\\mu (i.e., cos(\\xi)) out of bound\n");
            fprintf(stderr," mu,n_0,n,beta,cos_theta = ");
            fprintf(stderr,"%.16e %.16e %.16e %.16e %.16e\n",mu,n_0,n,beta,cos_theta);
            exit(-1);
#elif FLAG_ERR2 == ERR2_ON_SCREEN_CONT
            fprintf(stderr,"ERROR in alphanuStokes_integrand(), ");
            fprintf(stderr,"\\mu (i.e., cos(\\xi)) out of bound\n");
            fprintf(stderr," mu,n_0,n,beta,cos_theta = ");
            fprintf(stderr,"%.16e %.16e %.16e %.16e %.16e\n",mu,n_0,n,beta,cos_theta);
            if(mu < -1.) {
                  mu = -1.;
                  sinxi = 0.;
            } else {
                  mu = 1.;
                  sinxi = 0.;
            }
#elif FLAG_ERR2 == ERR2_TO_FILE_CONT
            fprintf(errorfixedpt,"ERROR in alphanuStokes_integrand(), ");
            fprintf(errorfixedpt," \\mu (i.e., cos(\\xi)) out of bound\n");
            fprintf(errorfixedpt," mu,n_0,n,beta,cos_theta = ");
            fprintf(errorfixedpt,"%.16e %.16e %.16e %.16e %.16e\n",
                                 mu,n_0,n,beta,cos_theta);
            fflush(errorfixedpt);
            if(mu < -1.) {
                  mu = -1.;
                  sinxi = 0.;
            } else {
                  mu = 1.;
                  sinxi = 0.;
            }
#elif FLAG_ERR2 == ERR2_FIX_CONT
            if(mu < -1.) {
                  mu = -1.;
                  sinxi = 0.;
            } else {
                  mu = 1.;
                  sinxi = 0.;
            }
#elif FLAG_ERR2 == ERR2_IGNORE
#else
  #error "FLAG_ERR2 value unknown"
#endif
      }

      /* Ramaty 1969 Eq.(8) (seems to be wrong, see Chanmugam & Dulk 81 Eq.(3) */
      arg_jn = n_0 * beta * sin_theta * sinxi;

      /* Bessel function routine: */
      d_jn = my_Bessel_dJ(n, arg_jn, &jn);

      tmp1 = jn   * (cos_theta - beta * mu) / sin_theta;
      tmp2 = d_jn * beta * sinxi;

      if(flag == 1) {
            tmp = tmp1*tmp1 + tmp2*tmp2;
      } else if(flag == 2) {
            tmp = tmp1*tmp1 - tmp2*tmp2;
      } else if(flag == 4) {
            tmp = -2*tmp1*tmp2;
      } else {
            fprintf(stderr,"ERROR in alphanuStokes_integrand(), ");
            fprintf(stderr,"flag=%d is not a valid value.\n",flag);
            exit(-1);
      }

      coeffDgam = gamma*gamma/C_c2; // coefficient of d/d\gamma term
      coeffDmu = (beta*cos_theta - mu) * gamma / (beta*beta); // coefficient of d/d\xi

      par.gamma = gamma;
      par.mu    = mu;
      par.tdl   = tdl;
      Fg.function = &ne_gsl_dg;
      Fm.function = &ne_gsl_dm;
      Fg.params = &par;
      Fm.params = &par;
      if(gamma-DGAMMA<1.) {  /* use forward method */
            gsl_deriv_forward(&Fg, gamma, DGAMMA, &d_ne_gam, &abserr_g);
      } else {               /* use central method */
            gsl_deriv_central(&Fg, gamma, DGAMMA, &d_ne_gam, &abserr_g);
      }
      if(mu-DMU<-1.) {       /* use forward method */
            gsl_deriv_forward(&Fm, mu, DMU, &d_ne_mu, &abserr_m);
      } else if(mu+DMU>1.) { /* use backward method */
            gsl_deriv_backward(&Fm, mu, DMU, &d_ne_mu, &abserr_m);
      } else {               /* use central method */
            gsl_deriv_central(&Fm, mu, DMU, &d_ne_mu, &abserr_m);
      }
      Df = coeffDgam * d_ne_gam + coeffDmu * d_ne_mu;

      return(tmp*Df);
}

/*============================================================================*/
/*============================================================================*/
/* Absorption coefficient calculated by using Gamma-n integration */
/* Can be one of ordinary mode (O), extraordinary mode (X),
 * Stokes parameters I, Q, V, or total */
double return_alphanu(int flagMODE,
               double nu, double rho, double Te, double B, double theta,
               size_t n_wksp, gsl_integration_workspace *wksp_g,
               gsl_integration_workspace *wksp_n, int flag_Tx_interp, void *Tx_list)
{
        double nrhoe,nuratio,nu_c,sin_theta,cos_theta,tdl,tdlinv;
        TYPE_PAR_n par_n;
        TYPE_PAR_jnu_I par_I;
        void *list = NULL;

        nrhoe = rho/(C_me+C_mp);
        sin_theta = sin(theta);
        cos_theta = cos(theta);
        nu_c = C_e*B/(C_2pi*C_me*C_c);
        nuratio = nu/nu_c;
        tdl = C_k * Te / C_me_c2;
        tdlinv = 1./tdl;

        /* assign parameters for integration */
        switch(flagMODE) {
                case TOTAL:
                        assign_par(&par_n,&par_I,alphanufunc_g,
                                   nuratio,nrhoe,tdl,B,sin_theta,cos_theta,
                                   n_wksp,wksp_g,wksp_n,0,list);
                        break;
                case MODE_O:
                        assign_par(&par_n,&par_I,alphanufuncO_g,
                                   nuratio,nrhoe,tdl,B,sin_theta,cos_theta,
                                   n_wksp,wksp_g,wksp_n,flag_Tx_interp,Tx_list);
                        break;
                case MODE_X:
                        assign_par(&par_n,&par_I,alphanufuncX_g,
                                   nuratio,nrhoe,tdl,B,sin_theta,cos_theta,
                                   n_wksp,wksp_g,wksp_n,flag_Tx_interp,Tx_list);
                        break;
                case STOKES_I:
                        assign_par(&par_n,&par_I,alphanufuncI_g,
                                   nuratio,nrhoe,tdl,B,sin_theta,cos_theta,
                                   n_wksp,wksp_g,wksp_n,0,list);
                        break;
                case STOKES_Q:
                        assign_par(&par_n,&par_I,alphanufuncQ_g,
                                   nuratio,nrhoe,tdl,B,sin_theta,cos_theta,
                                   n_wksp,wksp_g,wksp_n,0,list);
                        break;
                case STOKES_V:
                        assign_par(&par_n,&par_I,alphanufuncV_g,
                                   nuratio,nrhoe,tdl,B,sin_theta,cos_theta,
                                   n_wksp,wksp_g,wksp_n,0,list);
                        break;
                default:
                        fprintf(stderr,"ERR: invalid value of flagMODE in alphanu()\n");
                        fprintf(stderr,"     flagMODE = %d\n",flagMODE);
                        exit(91);
                        break;
        }

        /* do the summation (for low n) and integration */
        return(summation_and_integration(&par_n,&par_I)
               * extra_factor_alphanu(flagMODE,nrhoe,nu,cos_theta));

}

/*----------------------------------------------------------------------------*/

/* Wrapper of return_alphanu()
 * If flag EXTRAPOLATE is TRUE, and if theta is close to pi/2, then do
 *  extrapolation to get result. */
double alphanu(int flagMODE,
               double nu, double rho, double Te, double B, double theta,
               size_t n_wksp, gsl_integration_workspace *wksp_g,
               gsl_integration_workspace *wksp_n, int flag_Tx_interp, void *Tx_list)
{
#if (EXTRAPOLATE == FALSE)
        /* no extrapolation */
        return(return_alphanu(flagMODE,nu,rho,Te,B,theta,
                              n_wksp,wksp_g,wksp_n,flag_Tx_interp,Tx_list));
#elif (EXTRAPOLATE == TRUE)
        /* extrapolate near to Pi/2 */

        /* If theta is in (Pi/2 - DELTA_THETA , Pi/2 + DELTA_THETA), then 
         *  extrapolate.  Note that DELTA_THETA is assumed to be positive.  
         *  It should also be small (say, less than 0.0017453 rad or 
         *  0.1 deg). */
        /* Results are calculated at (Pi/2 + sign * DELTA_THETA) and
         *   (Pi/2 + sign * 2 * DELTA_THETA), and then extrapolate to theta.
         *   sign is -1 if theta < Pi/2, and +1 if theta > Pi/2 */

        int sign;
        double dtheta = theta - C_pi_2;
        if(dtheta < 0.) sign=-1;
        else sign=1;

        /* check whether we are close to pi/2 */
        if( fabs(dtheta) > DELTA_THETA ) {
                /* no extrapolation */
                return(return_alphanu(flagMODE,nu,rho,Te,B,theta,
                                      n_wksp,wksp_g,wksp_n,flag_Tx_interp,Tx_list));
        } else {
                /* extrapolation */
		double theta1 = C_pi_2 + sign * DELTA_THETA;
                double theta2 = C_pi_2 + sign * 2. * DELTA_THETA;
                double ans1 = return_alphanu(flagMODE,nu,rho,Te,B,theta1,
                                             n_wksp,wksp_g,wksp_n,
                                             flag_Tx_interp,Tx_list);
                double ans2 = return_alphanu(flagMODE,nu,rho,Te,B,theta2,
                                             n_wksp,wksp_g,wksp_n,
                                             flag_Tx_interp,Tx_list);
                return( (theta-theta1)*(ans2-ans1)/(theta2-theta1) + ans1 );
        }
#else
  #error not valid option for EXTRAPOLATE
#endif
}

/*============================================================================*/

/* angle-averaged j_\nu */
double jnu_average(int flagMODE,
                   double nu, double rho, double Te, double B, 
                   size_t n_wksp, gsl_integration_workspace *wksp_g,
                   gsl_integration_workspace *wksp_n)
{
        int i;
        double sum,dcostheta,costheta,theta;
        sum = 0.;
        dcostheta = 1./N_THETA_AVERAGE ;
        /* angle-averaged jnu = (1/2) Integrate[jnu(th) , {cos(th), -1, 1}]
         * Assume that jnu(th) = jnu(-th), therefore
         * angle-averaged jnu = Integrate[jnu(th) , {cos(th), 0, 1}] */
        for(i=0;i<N_THETA_AVERAGE;i++) {
                costheta = (i+0.5)*dcostheta;
                theta = acos(costheta);
                sum += jnu(flagMODE, nu, rho, Te, B, theta, N_WKSP, wksp_g, wksp_n);
        }
        return(sum*dcostheta);
}

/*============================================================================*/
/* vim:set ts=8 expandtab: */
