#include "decs.h"

/*============================================================================*/

/* example : isotropic nonthermal (power-law) distribution
 *
 * Synchrotron emissivity j_\nu and absorptivity \alpha_\nu, from frequency =
 *   1 \nu_c to 10^4 \nu_c, at 60 degree.
 *
 * Compare to approximate equations of j_\nu and \alpha_\nu
 *  in Leung, Gammie & Noble Eq.(58), (59).  Note that the approximate
 *  equations are valid when  \gamma_min^2 << \nu / \nu_c << \gamma_max^2 .
 * Also see Blumenthal & Gould, and also Rybicki & Lightman for details */

/* sanity check */
#if (ELECTRON_DIST != NONTHERMAL)
  #error This file assumes NONTHERMAL distribution is defined.  Set it in decs.h
#endif

/*============================================================================*/
int main(int argc, char *argv[])
{
        double nrhoe = 1.e5;
        double B = 100.;
        /* temperature in K */
        double Te = 1.e10; /* unused in this case */
        /* observer angle */
        double theta = (60.*C_pi/180.);
        /* number of data point, each for one frequency */
        int npt=21;
        int i;

        /* allocation of memory for gamma-n integration: */
        gsl_integration_workspace *wksp_g = gsl_integration_workspace_alloc(N_WKSP);
        gsl_integration_workspace *wksp_n = gsl_integration_workspace_alloc(N_WKSP);

        /* files for writing error */
#if FLAG_ERR1 == ERR1_TO_FILE_CONT || FLAG_ERR2 == ERR2_TO_FILE_CONT
        error_open_file();
#endif

        /* density */
        double rho = nrhoe*(C_me+C_mp);
        /* electron cyclotron frequency */
        double nu_c = C_e*B/(C_2pi*C_me*C_c);
        /* plasma frequency */
        double nu_p = C_e*sqrt(nrhoe/(C_me*C_pi));
        /* dimensionless electron temperature */
        //double tdl = C_k*Te/C_me_c2;

        fprintf(stdout,"# theta=%.8e B=%.8e  nrho_e=%.8e  nu_c=%.8e  nu_p=%.8e  index_p=%.8e  gmin=%.8e  gmax=%.8e  coeff_int=%.8e\n",
               theta,B,nrhoe,nu_c,nu_p,INDEX,GAM1,GAM2,COEFF_INT);
        fflush(stdout);

        /* min and max nu/nu_c */
        double nuratio1=1.e0,nuratio2=1.e4;
        for(i=0;i<npt;i++) {
                /* nu/nu_c */
                double nuratio = nuratio1 * pow(nuratio2/nuratio1,i/((double)(npt-1)));

                double nu = nuratio * nu_c;

                /* emissivity, without nrhoe */
                double j = jnu(TOTAL, nu, rho, Te, B, theta, 
                               N_WKSP, wksp_g, wksp_n) / nrhoe;

                /* Leung, Gammie & Noble, Eq.(58), without nrhoe */
                double j58 = (C_e2*nu_c/C_c)
                             *(pow(3.,0.5*INDEX)*(INDEX-1.)*sin(theta)
                              /(2.*(INDEX+1.)*(pow(GAM1,1.-INDEX)
                                              -pow(GAM2,1.-INDEX))))
                             *gsl_sf_gamma((3.*INDEX-1.)/12.)
                             *gsl_sf_gamma((3.*INDEX+19.)/12.)
                             *pow(nuratio/sin(theta),-0.5*(INDEX-1.));

                /* absorption coefficient, without nrhoe */
                void *list = NULL;
                double alpha = alphanu(TOTAL, nu, rho, Te, B, theta,
                                       N_WKSP, wksp_g, wksp_n, 0, list) / nrhoe;

                /* Leung, Gammie & Noble, Eq.(59), without nrhoe */
                double a59 = (C_e2/(nu*C_me*C_c))
                             *(pow(3.,0.5*(INDEX+1))*(INDEX-1.)*0.25
                              /(pow(GAM1,1.-INDEX)-pow(GAM2,1.-INDEX)))
                             *gsl_sf_gamma((3.*INDEX+2.)/12.)
                             *gsl_sf_gamma((3.*INDEX+22.)/12.)
                             *pow(nuratio/sin(theta),-0.5*(INDEX+2.));

                fprintf(stdout,"%.16e %.16e %.16e %.16e %.16e\n",
                       nuratio, j,  j58,  alpha, a59
                       );
                fflush(stdout);
        }

        /* free up memory after integration */
        gsl_integration_workspace_free(wksp_g);
        gsl_integration_workspace_free(wksp_n);

#if FLAG_ERR1 == ERR1_TO_FILE_CONT || FLAG_ERR2 == ERR2_TO_FILE_CONT
        error_close_file() ;
#endif

        return(0) ;
}
