#include "decs.h"

/*============================================================================*/

/* example : thermal distribution
 *
 * Synchrotron emissivity j_\nu and absorptivity \alpha_\nu
 *  at (k T_e)/(m_e c^2) = 10, from theta = 5 deg to 90 deg.  Calculation 
 *  at 90 deg is done by extrapolation.
 *
 * Compare to approximate equations of j_\nu
 *  in Leung, Gammie & Noble Eq.(72) */

/* sanity check */
#if (ELECTRON_DIST != THERMAL)
  #error This file assumes THERMAL distribution is defined.  Set it in decs.h
#endif

/*============================================================================*/

double funcRL(double x, void *para)
{
  return(gsl_sf_bessel_Knu(5./3,x));
}

int main(int argc, char *argv[])
{
        double nrhoe = 1.e5;
        double B = 30.;
        /* dimensionless temperature */
        double tdl = 10.;
        /* temperature in K */
        double Te = tdl*C_me_c2/C_k;
        /* number of data point, each for one frequency */
        int npt=35;
        int i;

        /* allocation of memory for gamma-n integration: */
        gsl_integration_workspace *wksp_g = gsl_integration_workspace_alloc(N_WKSP);
        gsl_integration_workspace *wksp_n = gsl_integration_workspace_alloc(N_WKSP);

        /* files for writing error */
#if FLAG_ERR1 == ERR1_TO_FILE_CONT || FLAG_ERR2 == ERR2_TO_FILE_CONT
        error_open_file();
#endif

        /* density */
        double rho = nrhoe*(C_me+C_mp);
        /* electron cyclotron frequency */
        double nu_c = C_e*B/(C_2pi*C_me*C_c);
        /* plasma frequency */
        double nu_p = C_e*sqrt(nrhoe/(C_me*C_pi));
        /* dimensionless electron temperature */
        //double tdl = C_k*Te/C_me_c2;

        /* nu/nu_c */
        double nuratio = 1.e2;

        /* frequency */
        double nu = nuratio * nu_c;

        fprintf(stdout,"# nuratio=%.8e  nu=%.8e  B=%.8e  nrho_e=%.8e  nu_c=%.8e  nu_p=%.8e\n",
               nuratio,nu,B,nrhoe,nu_c,nu_p);
        fflush(stdout);

        /* min and max angle (in deg) */
        double angle1=5.,angle2=90.;

        for(i=0;i<npt;i++) {
                /* observer angle in deg */
                double angle = angle1 + i*(angle2-angle1)/((double)(npt-1));

                /* observer angle in rad */
                double theta = angle*C_pi/180.;

                /* emissivity, without nrhoe */
                double j = jnu(TOTAL, nu, rho, Te, B, theta,
                               N_WKSP, wksp_g, wksp_n) / nrhoe;

                /* approximate emissivity in L,G&N, without nrhoe */
		double jT6 = jnu_T6(nu,rho,Te,B,theta) / nrhoe;

                /* absorptivity, without nrhoe */
                void *list = NULL;
                double alpha = alphanu(TOTAL, nu, rho, Te, B, theta,
                                       N_WKSP, wksp_g, wksp_n, 0, list) / nrhoe;

		fprintf(stdout,"%.16e %.16e %.16e %.16e\n",
		       angle, j, jT6, alpha
		       );
                fflush(stdout);
        }

        /* free up memory after integration */
        gsl_integration_workspace_free(wksp_g);
        gsl_integration_workspace_free(wksp_n);

#if FLAG_ERR1 == ERR1_TO_FILE_CONT || FLAG_ERR2 == ERR2_TO_FILE_CONT
        error_close_file() ;
#endif

        return(0) ;
}
